"""
Integrate Betty with `Webpack <https://webpack.js.org/>`_.

This module is internal.
"""

from __future__ import annotations

import asyncio
from pathlib import Path
from threading import Thread
from typing import TYPE_CHECKING, final, Self, ClassVar

from typing_extensions import override

from betty._npm import NpmRequirement, NpmUnavailable
from betty.html import CssProvider, JsProvider
from betty.jinja2 import Jinja2Provider, Filters, ContextVars
from betty.locale.localizable import static
from betty.os import copy_tree
from betty.plugin import ShorthandPluginBase
from betty.project.extension import Extension
from betty.project.extension.webpack import build
from betty.project.extension.webpack.build import EntryPointProvider
from betty.project.extension.webpack.jinja2.filter import FILTERS
from betty.project.generate import GenerateSiteEvent
from betty.requirement import Requirement, AllRequirements, RequirementError
from betty.typing import internal, private

if TYPE_CHECKING:
    from betty.job import Context
    from betty.project import Project
    from betty.event_dispatcher import EventHandlerRegistry
    from collections.abc import Sequence


async def _generate_assets(event: GenerateSiteEvent) -> None:
    project = event.project
    extensions = await project.extensions
    webpack = extensions[Webpack]
    build_directory_path = await webpack._generate_ensure_build_directory(
        job_context=event.job_context,
    )
    event.job_context._webpack_build_directory_path = build_directory_path  # type: ignore[attr-defined]
    await webpack._copy_build_directory(
        build_directory_path, project.configuration.www_directory_path
    )


@internal
@final
class Webpack(ShorthandPluginBase, Extension, CssProvider, JsProvider, Jinja2Provider):
    """
    Integrate Betty with `Webpack <https://webpack.js.org/>`_.
    """

    _plugin_id = "webpack"
    _plugin_label = static("Webpack")
    _requirement: ClassVar[Requirement | None] = None

    @private
    def __init__(
        self, project: Project, _public_css_path_prefix: str, _public_js_path: str
    ):
        super().__init__(project)
        self._public_css_path_prefix = _public_css_path_prefix
        self._public_js_path = _public_js_path

    @override
    @classmethod
    async def new_for_project(cls, project: Project) -> Self:
        url_generator = await project.url_generator
        return cls(
            project,
            url_generator.generate("betty-static:///css/"),
            url_generator.generate("betty-static:///js/webpack-entry-loader.js"),
        )

    @override
    def register_event_handlers(self, registry: EventHandlerRegistry) -> None:
        registry.add_handler(GenerateSiteEvent, _generate_assets)

    @override
    @classmethod
    async def requirement(cls) -> Requirement:
        if cls._requirement is None:
            cls._requirement = AllRequirements(
                await super().requirement(),
                await NpmRequirement.new(),
            )
        return cls._requirement

    @override
    @classmethod
    def assets_directory_path(cls) -> Path:
        return Path(__file__).parent / "assets"

    @override
    @property
    def public_css_paths(self) -> Sequence[str]:
        entry_points: Sequence[EntryPointProvider & Extension] = []

        def _target():
            entry_points.extend(asyncio.run(self._project_entry_point_providers()))

        thread = Thread(target=_target)
        thread.start()
        thread.join()
        return (
            f"{self._public_css_path_prefix}/webpack-vendor.css",
            *(
                f"{self._public_css_path_prefix}/webpack/{entry_point.plugin_id()}.css"
                for entry_point in entry_points
                if (
                    entry_point.webpack_entry_point_directory_path() / "main.scss"
                ).is_file()
            ),
        )

    @override
    @property
    def public_js_paths(self) -> Sequence[str]:
        return (self._public_js_path,)

    @override
    def new_context_vars(self) -> ContextVars:
        return {
            "webpack_js_entry_points": set(),
        }

    @override
    @property
    def filters(self) -> Filters:
        return FILTERS

    async def _project_entry_point_providers(
        self,
    ) -> Sequence[EntryPointProvider & Extension]:
        extensions = await self._project.extensions
        return [
            extension
            for extension in extensions.flatten()
            if isinstance(extension, EntryPointProvider)
        ]

    async def _new_builder(
        self,
        working_directory_path: Path,
        *,
        job_context: Context,
    ) -> build.Builder:
        return build.Builder(
            working_directory_path,
            await self._project_entry_point_providers(),
            self._project.configuration.debug,
            await self._project.renderer,
            self._project.configuration.root_path,
            job_context=job_context,
            localizer=await self._project.app.localizer,
        )

    async def _copy_build_directory(
        self,
        build_directory_path: Path,
        destination_directory_path: Path,
    ) -> None:
        await copy_tree(build_directory_path, destination_directory_path)

    async def _generate_ensure_build_directory(
        self,
        *,
        job_context: Context,
    ) -> Path:
        builder = await self._new_builder(
            self._project.app.binary_file_cache.with_scope("webpack").path,
            job_context=job_context,
        )
        try:
            # (Re)build the assets if `npm` is available.
            return await builder.build()
        except NpmUnavailable:
            raise RequirementError(await self.requirement()) from None



"""
Define and provide sequences of :py:class:`betty.config.Configuration` instances.
"""

from __future__ import annotations

from typing import (
    Generic,
    Iterable,
    MutableSequence,
    overload,
    Sequence,
    Iterator,
    TypeVar,
    TYPE_CHECKING,
)

from typing_extensions import override

from betty.assertion import assert_sequence
from betty.config import Configuration
from betty.config.collections import ConfigurationCollection

if TYPE_CHECKING:
    from betty.serde.dump import Dump, DumpSequence

_ConfigurationT = TypeVar("_ConfigurationT", bound=Configuration)


class ConfigurationSequence(
    ConfigurationCollection[int, _ConfigurationT], Generic[_ConfigurationT]
):
    """
    A sequence of configuration values.

    To test your own subclasses, use :py:class:`betty.test_utils.config.collections.sequence.ConfigurationSequenceTestBase`.
    """

    def __init__(
        self,
        configurations: Iterable[_ConfigurationT] | None = None,
    ):
        self._configurations: MutableSequence[_ConfigurationT] = []
        super().__init__(configurations)

    def __contains__(self, configuration: _ConfigurationT) -> bool:
        return configuration in self._configurations

    @override
    @overload
    def __getitem__(self, configuration_key: int) -> _ConfigurationT:
        pass

    @override
    @overload
    def __getitem__(self, configuration_key: slice) -> Sequence[_ConfigurationT]:
        pass

    @override
    def __getitem__(
        self, configuration_key: int | slice
    ) -> _ConfigurationT | Sequence[_ConfigurationT]:
        return self._configurations[configuration_key]

    @override
    def __iter__(self) -> Iterator[_ConfigurationT]:
        return (configuration for configuration in self._configurations)

    @override
    def keys(self) -> Iterator[int]:
        return iter(range(0, len(self._configurations)))

    @override
    def values(self) -> Iterator[_ConfigurationT]:
        yield from self._configurations

    @override
    def replace(self, *configurations: _ConfigurationT) -> None:
        self.assert_mutable()
        self.clear()
        self.append(*configurations)

    @override
    def load(self, dump: Dump) -> None:
        self.assert_mutable()
        self.replace(*assert_sequence(self._load_item)(dump))

    @override
    def dump(self) -> DumpSequence[Dump]:
        return [configuration.dump() for configuration in self._configurations]

    @override
    def prepend(self, *configurations: _ConfigurationT) -> None:
        self.assert_mutable()
        for configuration in configurations:
            self._pre_add(configuration)
            self._configurations.insert(0, configuration)

    @override
    def append(self, *configurations: _ConfigurationT) -> None:
        self.assert_mutable()
        for configuration in configurations:
            self._pre_add(configuration)
            self._configurations.append(configuration)

    @override
    def insert(self, index: int, *configurations: _ConfigurationT) -> None:
        self.assert_mutable()
        for configuration in reversed(configurations):
            self._pre_add(configuration)
            self._configurations.insert(index, configuration)



"""
Test utilities for :py:mod:`betty.ancestry.date`.
"""

from betty.ancestry.date import HasDate


class DummyHasDate(HasDate):
    """
    A dummy :py:class:`betty.ancestry.date.HasDate` implementation.
    """

    pass



"""
Entity collections.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from reprlib import recursive_repr
from typing import (
    Generic,
    Iterator,
    overload,
    Any,
    Iterable,
    cast,
    TypeVar,
    TYPE_CHECKING,
    Self,
)

from typing_extensions import override

from betty.functools import unique
from betty.model import Entity
from betty.mutability import Mutable
from betty.repr import repr_instance

if TYPE_CHECKING:
    from betty.plugin import PluginIdToTypeMapping
    from betty.machine_name import MachineName
    from collections.abc import Sequence, MutableSequence, MutableMapping, AsyncIterator

_EntityT = TypeVar("_EntityT", bound=Entity)
_TargetT = TypeVar("_TargetT")


class EntityCollection(Mutable, Generic[_TargetT], ABC):
    """
    Provide a collection of entities.

    To test your own subclasses, use :py:class:`betty.test_utils.model.collections.EntityCollectionTestBase`.
    """

    __slots__ = ()

    def _on_add(self, *entities: _TargetT & Entity) -> None:
        pass

    def _on_remove(self, *entities: _TargetT & Entity) -> None:
        pass

    @override
    def get_mutable_instances(self) -> Iterable[Mutable]:
        return self

    @property
    def view(self) -> Sequence[_TargetT & Entity]:
        """
        A view of the entities at the time of calling.
        """
        return [*self]

    @abstractmethod
    def add(self, *entities: _TargetT & Entity) -> None:
        """
        Add the given entities.
        """
        pass

    @abstractmethod
    def remove(self, *entities: _TargetT & Entity) -> None:
        """
        Remove the given entities.
        """
        pass

    def replace(self, *entities: _TargetT & Entity) -> None:
        """
        Replace all entities with the given ones.
        """
        self.remove(*(entity for entity in self if entity not in entities))
        self.add(*entities)

    @abstractmethod
    def clear(self) -> None:
        """
        Clear all entities from the collection.
        """
        pass

    @abstractmethod
    def __iter__(self) -> Iterator[_TargetT & Entity]:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @overload
    def __getitem__(self, index: int) -> _TargetT & Entity:
        pass

    @overload
    def __getitem__(self, indices: slice) -> Sequence[_TargetT & Entity]:
        pass

    @abstractmethod
    def __getitem__(
        self, key: int | slice
    ) -> _TargetT & Entity | Sequence[_TargetT & Entity]:
        pass

    @abstractmethod
    def __delitem__(self, key: _TargetT & Entity) -> None:
        pass

    @abstractmethod
    def __contains__(self, value: Any) -> bool:
        pass

    def _known(self, *entities: _TargetT & Entity) -> Iterable[_TargetT & Entity]:
        for entity in unique(entities):
            if entity in self:
                yield entity

    def _unknown(self, *entities: _TargetT & Entity) -> Iterable[_TargetT & Entity]:
        for entity in unique(entities):
            if entity not in self:
                yield entity


_EntityCollectionT = TypeVar("_EntityCollectionT", bound=EntityCollection[_EntityT])


class SingleTypeEntityCollection(Generic[_TargetT], EntityCollection[_TargetT]):
    """
    Collect entities of a single type.
    """

    __slots__ = "_entities", "_target_type"

    def __init__(self, target_type: type[_TargetT], *entities: _TargetT & Entity):
        super().__init__()
        self._entities: MutableSequence[_TargetT & Entity] = [*entities]
        self._target_type = target_type

    @override  # type: ignore[callable-functiontype]
    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, target_type=self._target_type, length=len(self))

    @override
    def add(self, *entities: _TargetT & Entity) -> None:
        added_entities = [*self._unknown(*entities)]
        for entity in added_entities:
            self._entities.append(entity)
        if added_entities:
            self._on_add(*added_entities)

    @override
    def remove(self, *entities: _TargetT & Entity) -> None:
        removed_entities = [*self._known(*entities)]
        for entity in removed_entities:
            self._entities.remove(entity)
        if removed_entities:
            self._on_remove(*removed_entities)

    @override
    def clear(self) -> None:
        self.remove(*self)

    @override
    def __iter__(self) -> Iterator[_TargetT & Entity]:
        return self._entities.__iter__()

    @override
    def __len__(self) -> int:
        return len(self._entities)

    @overload
    def __getitem__(self, index: int) -> _TargetT & Entity:
        pass

    @overload
    def __getitem__(self, indices: slice) -> Sequence[_TargetT & Entity]:
        pass

    @overload
    def __getitem__(self, entity_id: str) -> _TargetT & Entity:
        pass

    @override
    def __getitem__(
        self, key: int | slice | str
    ) -> _TargetT & Entity | Sequence[_TargetT & Entity]:
        if isinstance(key, int):
            return self._getitem_by_index(key)
        if isinstance(key, slice):
            return self._getitem_by_indices(key)
        return self._getitem_by_entity_id(key)

    def _getitem_by_index(self, index: int) -> _TargetT & Entity:
        return self._entities[index]

    def _getitem_by_indices(self, indices: slice) -> Sequence[_TargetT & Entity]:
        return self.view[indices]

    def _getitem_by_entity_id(self, entity_id: str) -> _TargetT & Entity:
        for entity in self._entities:
            if entity_id == entity.id:
                return entity
        raise KeyError(
            f'Cannot find a {self._target_type} entity with ID "{entity_id}".'
        )

    @override
    def __delitem__(self, key: str | _TargetT & Entity) -> None:
        if isinstance(key, self._target_type):
            return self._delitem_by_entity(cast("_TargetT & Entity", key))
        if isinstance(key, str):
            return self._delitem_by_entity_id(key)
        raise TypeError(f"Cannot find entities by {repr(key)}.")

    def _delitem_by_entity(self, entity: _TargetT & Entity) -> None:
        self.remove(entity)

    def _delitem_by_entity_id(self, entity_id: str) -> None:
        for entity in self._entities:
            if entity_id == entity.id:
                self.remove(entity)
                return

    @override
    def __contains__(self, value: Any) -> bool:
        if isinstance(value, self._target_type):
            return self._contains_by_entity(cast("_TargetT & Entity", value))
        if isinstance(value, str):
            return self._contains_by_entity_id(value)
        return False

    def _contains_by_entity(self, other_entity: _TargetT & Entity) -> bool:
        return any(other_entity is entity for entity in self._entities)

    def _contains_by_entity_id(self, entity_id: str) -> bool:
        return any(entity.id == entity_id for entity in self._entities)


class MultipleTypesEntityCollection(Generic[_TargetT], EntityCollection[_TargetT]):
    """
    Collect entities of multiple types.
    """

    __slots__ = ("_collections", "_entity_type_id_to_type_mapping")

    def __init__(
        self,
        *entities: _TargetT & Entity,
        entity_type_id_to_type_mapping: PluginIdToTypeMapping[Entity],
    ):
        super().__init__()
        self._entity_type_id_to_type_mapping = entity_type_id_to_type_mapping
        self._collections: MutableMapping[
            type[Entity], SingleTypeEntityCollection[Entity]
        ] = {}
        self.add(*entities)

    @classmethod
    async def new(cls, *entities: _TargetT & Entity) -> Self:
        """
        Create a new instance.
        """
        from betty.model import ENTITY_TYPE_REPOSITORY

        return cls(
            *entities,
            entity_type_id_to_type_mapping=await ENTITY_TYPE_REPOSITORY.mapping(),
        )

    @override  # type: ignore[callable-functiontype]
    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(
            self,
            entity_types=", ".join(
                entity_type.plugin_id() for entity_type in self._collections
            ),
            length=len(self),
        )

    def _get_collection(
        self, entity_type: type[_EntityT]
    ) -> SingleTypeEntityCollection[_EntityT]:
        assert issubclass(entity_type, Entity), f"{entity_type} is not an entity type."
        try:
            return cast(
                "SingleTypeEntityCollection[_EntityT]", self._collections[entity_type]
            )
        except KeyError:
            self._collections[entity_type] = SingleTypeEntityCollection(entity_type)
            return cast(
                "SingleTypeEntityCollection[_EntityT]", self._collections[entity_type]
            )

    @overload
    def __getitem__(self, index: int) -> _TargetT & Entity:
        pass

    @overload
    def __getitem__(self, indices: slice) -> Sequence[_TargetT & Entity]:
        pass

    @overload
    def __getitem__(
        self, entity_type_id: MachineName
    ) -> SingleTypeEntityCollection[Entity]:
        pass

    @overload
    def __getitem__(
        self, entity_type: type[_EntityT]
    ) -> SingleTypeEntityCollection[_EntityT]:
        pass

    @override
    def __getitem__(
        self,
        key: int | slice | str | type[_EntityT],
    ) -> (
        _TargetT & Entity
        | SingleTypeEntityCollection[Entity]
        | SingleTypeEntityCollection[_EntityT]
        | Sequence[_TargetT & Entity]
    ):
        if isinstance(key, int):
            return self._getitem_by_index(key)
        if isinstance(key, slice):
            return self._getitem_by_indices(key)
        if isinstance(key, str):
            return self._getitem_by_entity_type_id(key)
        return self._getitem_by_entity_type(key)

    def _getitem_by_entity_type(
        self, entity_type: type[_EntityT]
    ) -> SingleTypeEntityCollection[_EntityT]:
        return self._get_collection(entity_type)

    def _getitem_by_entity_type_id(
        self, entity_type_id: MachineName
    ) -> SingleTypeEntityCollection[Entity]:
        return self._get_collection(
            self._entity_type_id_to_type_mapping[entity_type_id]
        )

    def _getitem_by_index(self, index: int) -> _TargetT & Entity:
        return self.view[index]

    def _getitem_by_indices(self, indices: slice) -> Sequence[_TargetT & Entity]:
        return self.view[indices]

    @override
    def __delitem__(
        self, key: str | type[_TargetT & Entity] | _TargetT & Entity
    ) -> None:
        if isinstance(key, type):
            return self._delitem_by_entity_type(
                key,
            )
        if isinstance(key, Entity):
            return self._delitem_by_entity(
                key,  # type: ignore[arg-type]
            )
        return self._delitem_by_entity_type_id(key)

    def _delitem_by_entity_type(self, entity_type: type[_TargetT & Entity]) -> None:
        removed_entities = [*self._get_collection(entity_type)]
        self._get_collection(entity_type).clear()
        if removed_entities:
            self._on_remove(*removed_entities)

    def _delitem_by_entity(self, entity: _TargetT & Entity) -> None:
        self.remove(entity)

    def _delitem_by_entity_type_id(self, entity_type_id: MachineName) -> None:
        self._delitem_by_entity_type(
            self._entity_type_id_to_type_mapping[entity_type_id]  # type: ignore [arg-type]
        )

    @override
    def __iter__(self) -> Iterator[_TargetT & Entity]:
        for collection in self._collections.values():
            for entity in collection:
                yield cast("_TargetT & Entity", entity)

    @override
    def __len__(self) -> int:
        return sum(map(len, self._collections.values()))

    @override
    def __contains__(self, value: Any) -> bool:
        if isinstance(value, Entity):
            return self._contains_by_entity(value)
        return False

    def _contains_by_entity(self, other_entity: Any) -> bool:
        return any(other_entity is entity for entity in self)

    @override
    def add(self, *entities: _TargetT & Entity) -> None:
        added_entities = [*self._unknown(*entities)]
        for entity in added_entities:
            self[entity.type].add(entity)
        if added_entities:
            self._on_add(*added_entities)

    @override
    def remove(self, *entities: _TargetT & Entity) -> None:
        removed_entities = [*self._known(*entities)]
        for entity in removed_entities:
            self[entity.type].remove(entity)
        if removed_entities:
            self._on_remove(*removed_entities)

    @override
    def clear(self) -> None:
        removed_entities = (*self,)
        for collection in self._collections.values():
            collection.clear()
        if removed_entities:
            self._on_remove(*removed_entities)


@asynccontextmanager
async def record_added(
    entities: EntityCollection[_EntityT],
) -> AsyncIterator[MultipleTypesEntityCollection[_EntityT]]:
    """
    Record all entities that are added to a collection.
    """
    original = [*entities]
    added = await MultipleTypesEntityCollection[_EntityT].new()
    yield added
    added.add(*[entity for entity in entities if entity not in original])



"""
Provide caching that stores cache items in volatile memory.
"""

from __future__ import annotations

import multiprocessing
from collections.abc import MutableMapping, Sequence
from typing import TypeAlias, Generic, Self, TypeVar, final, TYPE_CHECKING

from typing_extensions import override

from betty.cache import CacheItem
from betty.cache._base import _CommonCacheBase, _StaticCacheItem, _CommonCacheBaseState
from betty.typing import processsafe

if TYPE_CHECKING:
    from betty.concurrent import Ledger, AsynchronizedLock
    from multiprocessing.managers import SyncManager

_CacheItemValueContraT = TypeVar("_CacheItemValueContraT", contravariant=True)

_MemoryCacheStore: TypeAlias = MutableMapping[
    tuple[str, ...],
    "CacheItem[_CacheItemValueContraT] | None | _MemoryCacheStore[_CacheItemValueContraT]",
]


@final
class _MemoryCacheState(
    Generic[_CacheItemValueContraT],
    _CommonCacheBaseState["MemoryCache[_CacheItemValueContraT]"],
):
    def __init__(
        self,
        cache_lock: AsynchronizedLock,
        cache_item_lock_ledger: Ledger,
        store: _MemoryCacheStore[_CacheItemValueContraT],
    ):
        super().__init__(cache_lock, cache_item_lock_ledger)
        self.store = store


@final
@processsafe
class MemoryCache(
    _CommonCacheBase[_CacheItemValueContraT], Generic[_CacheItemValueContraT]
):
    """
    Provide a cache that stores cache items in volatile memory.
    """

    def __init__(
        self,
        *,
        scopes: Sequence[str] | None = None,
        manager: SyncManager | _MemoryCacheState[_CacheItemValueContraT] | None = None,
    ):
        super().__init__(scopes=scopes, manager=manager)
        if isinstance(manager, _MemoryCacheState):
            self._store = manager.store
        else:
            self._store = multiprocessing.Manager().dict()

    @override
    def with_scope(self, scope: str) -> Self:
        return type(self)(
            scopes=(*self._scopes, scope),
            manager=_MemoryCacheState[_CacheItemValueContraT](
                self._cache_lock, self._cache_item_lock_ledger, self._store
            ),
        )

    def _cache_item_key(self, cache_item_id: str) -> tuple[str, ...]:
        return *self._scopes, cache_item_id

    @override
    async def _get(
        self, cache_item_id: str
    ) -> CacheItem[_CacheItemValueContraT] | None:
        cache_item = self._store.get(self._cache_item_key(cache_item_id), None)
        if isinstance(cache_item, CacheItem):
            return cache_item
        return None

    @override
    async def _set(
        self,
        cache_item_id: str,
        value: _CacheItemValueContraT,
        *,
        modified: int | float | None = None,
    ) -> None:
        self._store[self._cache_item_key(cache_item_id)] = _StaticCacheItem(
            value, modified
        )

    @override
    async def _delete(self, cache_item_id: str) -> None:
        self._store.pop(self._cache_item_key(cache_item_id), None)

    @override
    async def _clear(self) -> None:
        self._store.clear()



"""
Provide the HTML API, for generating HTML pages.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence, MutableMapping
from inspect import getmembers
from typing import (
    MutableSequence,
    final,
    Self,
    cast,
    Generic,
    TypeVar,
    TypedDict,
    NotRequired,
    Unpack,
    Any,
)

from typing_extensions import override, overload

from betty.string import (
    kebab_case_to_lower_camel_case,
    snake_case_to_kebab_case,
    kebab_case_to_snake_case,
)

_AttributeGetT = TypeVar("_AttributeGetT")
_AttributeSetT = TypeVar("_AttributeSetT")


class _Attribute(Generic[_AttributeGetT, _AttributeSetT], ABC):
    def __init__(self, html_name: str):
        self._html_name = html_name
        self._attr_name = f"_{kebab_case_to_lower_camel_case(html_name)}"

    @overload
    def __get__(self, instance: None, owner: type[Attributes]) -> Self:
        pass

    @overload
    def __get__(self, instance: Attributes, owner: type[Attributes]) -> _AttributeGetT:
        pass

    def __get__(
        self, instance: Attributes | None, owner: type[Attributes]
    ) -> _AttributeGetT | Self:
        if instance is None:
            return self  # type: ignore[return-value]
        return self.get(instance)

    def get(self, instance: Attributes) -> _AttributeGetT:
        try:
            return cast("_AttributeGetT", getattr(instance, self._attr_name))
        except AttributeError:
            value = self._new_default()
            setattr(instance, self._attr_name, value)
            return value

    def __set__(self, instance: Attributes, value: _AttributeSetT) -> None:
        self.set(instance, value)

    @abstractmethod
    def set(self, instance: Attributes, value: _AttributeSetT) -> None:
        pass

    def setdefault(self, instance: Attributes, value: _AttributeSetT) -> None:
        if getattr(instance, self._attr_name, None):
            return
        self.set(instance, value)

    @abstractmethod
    def _new_default(self) -> _AttributeGetT:
        pass

    def format(self, instance: Attributes) -> str:
        """
        Format the attribute to a string.
        """
        value = self.get(instance)
        if value:
            return self._format_value(value)
        return ""

    @abstractmethod
    def _format_value(self, value: _AttributeGetT) -> str:
        pass


class _BooleanAttribute(_Attribute[bool, bool]):
    @override
    def set(self, instance: Attributes, value: bool) -> None:
        setattr(instance, self._attr_name, value)

    @override
    def _format_value(self, value: bool) -> str:
        return self._html_name

    @override
    def _new_default(self) -> bool:
        return False


class _StringAttribute(_Attribute[str, str]):
    @override
    def set(self, instance: Attributes, value: str) -> None:
        setattr(instance, self._attr_name, value)

    @override
    def _format_value(self, value: str) -> str:
        return f'{self._html_name}="{value}"'

    @override
    def _new_default(self) -> str:
        return ""


class _MultipleStringAttribute(_Attribute[MutableSequence[str], Sequence[str]]):
    def __init__(self, html_name: str, separator: str = " "):
        super().__init__(html_name)
        self._separator = separator

    @override
    def set(self, instance: Attributes, value: Sequence[str]) -> None:
        sequence = self.get(instance)
        sequence.clear()
        sequence.extend(value)

    @override
    def _format_value(self, value: Sequence[str]) -> str:
        return f'{self._html_name}="{self._separator.join(value)}"'

    @override
    def _new_default(self) -> MutableSequence[str]:
        return []


class _BooleanOrStringAttribute(_Attribute[bool | str, bool | str]):
    @override
    def set(self, instance: Attributes, value: bool | str) -> None:
        setattr(instance, self._attr_name, value)

    @override
    def _format_value(self, value: bool | str) -> str:
        if isinstance(value, bool):
            return self._html_name
        return f'{self._html_name}="{value}"'

    @override
    def _new_default(self) -> bool | str:
        return False


class _AttributesKwargs(TypedDict):
    html_accept: NotRequired[Sequence[str]]
    html_accept_charset: NotRequired[str]
    html_accesskey: NotRequired[str]
    html_action: NotRequired[str]
    html_allow: NotRequired[str]
    html_alt: NotRequired[str]
    html_as: NotRequired[str]
    html_async: NotRequired[bool]
    html_autocapitalize: NotRequired[str]
    html_autocomplete: NotRequired[str]
    html_autoplay: NotRequired[bool]
    html_capture: NotRequired[str]
    html_charset: NotRequired[str]
    html_checked: NotRequired[bool]
    html_cite: NotRequired[str]
    html_class: NotRequired[Sequence[str]]
    html_cols: NotRequired[str]
    html_colspan: NotRequired[str]
    html_content: NotRequired[str]
    html_contenteditable: NotRequired[str]
    html_controls: NotRequired[bool]
    html_coords: NotRequired[str]
    html_crossorigin: NotRequired[str]
    html_data: NotRequired[str]
    html_datetime: NotRequired[str]
    html_decoding: NotRequired[str]
    html_default: NotRequired[bool]
    html_defer: NotRequired[bool]
    html_dir: NotRequired[str]
    html_dirname: NotRequired[str]
    html_disabled: NotRequired[bool]
    html_download: NotRequired[bool | str]
    html_draggable: NotRequired[str]
    html_enctype: NotRequired[str]
    html_enterkeyhint: NotRequired[str]
    html_for: NotRequired[str]
    html_formaction: NotRequired[str]
    html_formenctype: NotRequired[str]
    html_formmethod: NotRequired[str]
    html_formnovalidate: NotRequired[bool]
    html_formtarget: NotRequired[str]
    html_headers: NotRequired[Sequence[str]]
    html_height: NotRequired[str]
    html_hidden: NotRequired[str]
    html_high: NotRequired[str]
    html_href: NotRequired[str]
    html_hreflang: NotRequired[str]
    html_http_equiv: NotRequired[str]
    html_id: NotRequired[str]
    html_integrity: NotRequired[str]
    html_inputmode: NotRequired[str]
    html_ismap: NotRequired[bool]
    html_itemprop: NotRequired[str]
    html_kind: NotRequired[str]
    html_label: NotRequired[str]
    html_lang: NotRequired[str]
    html_loading: NotRequired[str]
    html_list: NotRequired[str]
    html_loop: NotRequired[bool]
    html_low: NotRequired[str]
    html_max: NotRequired[str]
    html_maxlength: NotRequired[str]
    html_minlength: NotRequired[str]
    html_media: NotRequired[str]
    html_method: NotRequired[str]
    html_min: NotRequired[str]
    html_multiple: NotRequired[bool]
    html_muted: NotRequired[bool]
    html_name: NotRequired[str]
    html_novalidate: NotRequired[bool]
    html_open: NotRequired[bool]
    html_optimum: NotRequired[str]
    html_pattern: NotRequired[str]
    html_ping: NotRequired[Sequence[str]]
    html_placeholder: NotRequired[str]
    html_playsinline: NotRequired[bool]
    html_poster: NotRequired[str]
    html_preload: NotRequired[str]
    html_readonly: NotRequired[bool]
    html_referrerpolicy: NotRequired[str]
    html_rel: NotRequired[str]
    html_required: NotRequired[bool]
    html_reversed: NotRequired[bool]
    html_role: NotRequired[str]
    html_rows: NotRequired[str]
    html_rowspan: NotRequired[str]
    html_sandbox: NotRequired[Sequence[str]]
    html_scope: NotRequired[str]
    html_selected: NotRequired[bool]
    html_shape: NotRequired[str]
    html_size: NotRequired[str]
    html_sizes: NotRequired[str]
    html_slot: NotRequired[str]
    html_span: NotRequired[str]
    html_spellcheck: NotRequired[str]
    html_src: NotRequired[str]
    html_srcdoc: NotRequired[str]
    html_srclang: NotRequired[str]
    html_srcset: NotRequired[Sequence[str]]
    html_start: NotRequired[str]
    html_step: NotRequired[str]
    html_style: NotRequired[str]
    html_tabindex: NotRequired[str]
    html_target: NotRequired[str]
    html_title: NotRequired[str]
    html_translate: NotRequired[str]
    html_type: NotRequired[str]
    html_usemap: NotRequired[str]
    html_value: NotRequired[str]
    html_width: NotRequired[str]
    html_wrap: NotRequired[str]


@final
class Attributes:
    """
    Manage attributes for an HTML element.
    """

    # Based on https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes.
    html_accept = _MultipleStringAttribute("accept", ", ")
    html_accept_charset = _StringAttribute("accept-charset")
    html_accesskey = _StringAttribute("accesskey")
    html_action = _StringAttribute("action")
    html_allow = _StringAttribute("allow")
    html_alt = _StringAttribute("alt")
    html_as = _StringAttribute("as")
    html_async = _BooleanAttribute("async")
    html_autocapitalize = _StringAttribute("autocapitalize")
    html_autocomplete = _StringAttribute("autocomplete")
    html_autoplay = _BooleanAttribute("autoplay")
    html_capture = _StringAttribute("capture")
    html_charset = _StringAttribute("charset")
    html_checked = _BooleanAttribute("checked")
    html_cite = _StringAttribute("cite")
    html_class = _MultipleStringAttribute("class")
    html_cols = _StringAttribute("cols")
    html_colspan = _StringAttribute("colspan")
    html_content = _StringAttribute("content")
    html_contenteditable = _StringAttribute("contenteditable")
    html_controls = _BooleanAttribute("controls")
    html_coords = _StringAttribute("coords")
    html_crossorigin = _StringAttribute("crossorigin")
    # @todo "csp" is not yet documented by MDN.
    html_data = _StringAttribute("data")
    html_datetime = _StringAttribute("datetime")
    html_decoding = _StringAttribute("decoding")
    html_default = _BooleanAttribute("default")
    html_defer = _BooleanAttribute("defer")
    html_dir = _StringAttribute("dir")
    html_dirname = _StringAttribute("dirname")
    html_disabled = _BooleanAttribute("disabled")
    html_download = _BooleanOrStringAttribute("download")
    html_draggable = _StringAttribute("draggable")
    html_enctype = _StringAttribute("enctype")
    html_enterkeyhint = _StringAttribute("enterkeyhint")
    html_for = _StringAttribute("for")
    # @todo "form" is not yet documented by MDN.
    html_formaction = _StringAttribute("formaction")
    html_formenctype = _StringAttribute("formenctype")
    html_formmethod = _StringAttribute("formmethod")
    html_formnovalidate = _BooleanAttribute("formnovalidate")
    html_formtarget = _StringAttribute("formtarget")
    html_headers = _MultipleStringAttribute("headers")
    html_height = _StringAttribute("height")
    html_hidden = _StringAttribute("hidden")
    html_high = _StringAttribute("high")
    html_href = _StringAttribute("href")
    html_hreflang = _StringAttribute("hreflang")
    html_http_equiv = _StringAttribute("http-equiv")
    html_id = _StringAttribute("id")
    html_integrity = _StringAttribute("integrity")
    html_inputmode = _StringAttribute("inputmode")
    html_ismap = _BooleanAttribute("ismap")
    html_itemprop = _StringAttribute("itemprop")
    html_kind = _StringAttribute("kind")
    html_label = _StringAttribute("label")
    html_lang = _StringAttribute("lang")
    html_loading = _StringAttribute("loading")
    html_list = _StringAttribute("list")
    html_loop = _BooleanAttribute("loop")
    html_low = _StringAttribute("low")
    html_max = _StringAttribute("max")
    html_maxlength = _StringAttribute("maxlength")
    html_minlength = _StringAttribute("minlength")
    html_media = _StringAttribute("media")
    html_method = _StringAttribute("method")
    html_min = _StringAttribute("min")
    html_multiple = _BooleanAttribute("multiple")
    html_muted = _BooleanAttribute("muted")
    html_name = _StringAttribute("name")
    html_novalidate = _BooleanAttribute("novalidate")
    html_open = _BooleanAttribute("open")
    html_optimum = _StringAttribute("optimum")
    html_pattern = _StringAttribute("pattern")
    html_ping = _MultipleStringAttribute("ping")
    html_placeholder = _StringAttribute("placeholder")
    html_playsinline = _BooleanAttribute("playsinline")
    html_poster = _StringAttribute("poster")
    html_preload = _StringAttribute("preload")
    html_readonly = _BooleanAttribute("readonly")
    html_referrerpolicy = _StringAttribute("referrerpolicy")
    html_rel = _StringAttribute("rel")
    html_required = _BooleanAttribute("required")
    html_reversed = _BooleanAttribute("reversed")
    html_role = _StringAttribute("role")
    html_rows = _StringAttribute("rows")
    html_rowspan = _StringAttribute("rowspan")
    html_sandbox = _MultipleStringAttribute("sandbox")
    html_scope = _StringAttribute("scope")
    html_selected = _BooleanAttribute("selected")
    html_shape = _StringAttribute("shape")
    html_size = _StringAttribute("size")
    html_sizes = _StringAttribute("sizes")
    html_slot = _StringAttribute("slot")
    html_span = _StringAttribute("span")
    html_spellcheck = _StringAttribute("spellcheck")
    html_src = _StringAttribute("src")
    html_srcdoc = _StringAttribute("srcdoc")
    html_srclang = _StringAttribute("srclang")
    html_srcset = _MultipleStringAttribute("srcset", ", ")
    html_start = _StringAttribute("start")
    html_step = _StringAttribute("step")
    html_style = _StringAttribute("style")
    html_tabindex = _StringAttribute("tabindex")
    html_target = _StringAttribute("target")
    html_title = _StringAttribute("title")
    html_translate = _StringAttribute("translate")
    html_type = _StringAttribute("type")
    html_usemap = _StringAttribute("usemap")
    html_value = _StringAttribute("value")
    html_width = _StringAttribute("width")
    html_wrap = _StringAttribute("wrap")

    def __init__(self, **kwargs: Unpack[_AttributesKwargs]):
        self._data_attributes: MutableMapping[str, str] = {}
        self.set(**kwargs)

    def _get_attribute(self, attr_name: str) -> _Attribute[Any, Any]:
        return cast("_Attribute[Any, Any]", getattr(type(self), attr_name))

    def set(self, **attributes: Unpack[_AttributesKwargs]) -> None:
        """
        Set values for the given HTML attributes.
        """
        for attribute_name, attribute_value in attributes.items():
            self._get_attribute(attribute_name).set(self, attribute_value)

    def setdefault(self, **attributes: Unpack[_AttributesKwargs]) -> None:
        """
        Set values for the given HTML attributes, but only for those attributes that do not already have a value set.
        """
        for attribute_name, attribute_value in attributes.items():
            self._get_attribute(attribute_name).setdefault(self, attribute_value)

    def set_data(self, **attributes: str) -> None:
        """
        Set values for the given HTML data attributes.
        """
        self._data_attributes.update(attributes)

    def get_data(self, attribute_name: str) -> str | None:
        """
        Get the value for the given HTML data attribute.
        """
        try:
            return self._data_attributes[kebab_case_to_snake_case(attribute_name)]
        except KeyError:
            return None

    def format(self) -> str:
        """
        Format the HTML attributes to a string.
        """
        return " ".join(
            (
                *(
                    formatted_attribute
                    for formatted_attribute in (
                        value.format(self)
                        for _name, value in getmembers(type(self))
                        if isinstance(value, _Attribute)
                    )
                    if formatted_attribute
                ),
                *(
                    f'data-{snake_case_to_kebab_case(attribute_name)}="{attribute_value}"'
                    for attribute_name, attribute_value in self._data_attributes.items()
                ),
            )
        )

    def __str__(self) -> str:
        return self.format()

    def __html__(self) -> str:
        return self.format()



"""
Date types with dates.
"""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from typing_extensions import override

from betty.date import Datey, Date, DateySchema
from betty.json.linked_data import LinkedDataDumpableJsonLdObject, JsonLdObject
from betty.privacy import is_public

if TYPE_CHECKING:
    from betty.serde.dump import DumpMapping, Dump
    from betty.project import Project


class HasDate(LinkedDataDumpableJsonLdObject):
    """
    A resource with date information.
    """

    def __init__(
        self,
        *args: Any,
        date: Datey | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.date = date

    def dated_linked_data_contexts(self) -> tuple[str | None, str | None, str | None]:
        """
        Get the JSON-LD context term definition IRIs for the possible dates.

        :returns: A 3-tuple with the IRI for a single date, a start date, and an end date, respectively.
        """
        return None, None, None

    @override
    async def dump_linked_data(self, project: Project) -> DumpMapping[Dump]:
        dump = await super().dump_linked_data(project)
        if self.date and is_public(self):
            (
                schema_org_date_definition,
                schema_org_start_date_definition,
                schema_org_end_date_definition,
            ) = self.dated_linked_data_contexts()
            if isinstance(self.date, Date):
                dump["date"] = await self.date.dump_linked_data(
                    project, schema_org_date_definition
                )
            else:
                dump["date"] = await self.date.dump_linked_data(
                    project,
                    schema_org_start_date_definition,
                    schema_org_end_date_definition,
                )
        return dump

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property("date", await DateySchema.new(), False)
        return schema



"""
Provide configuration for the Wikipedia extension.
"""

from typing_extensions import override

from betty.assertion import (
    OptionalField,
    assert_record,
    assert_bool,
    assert_setattr,
)
from betty.config import Configuration
from betty.serde.dump import Dump, DumpMapping


class WikipediaConfiguration(Configuration):
    """
    Provides configuration for the :py:class:`betty.project.extension.wikipedia.Wikipedia` extension.
    """

    def __init__(self):
        super().__init__()
        self._populate_images = True

    @property
    def populate_images(self) -> bool:
        """
        Whether to populate entities with Wikimedia images after loading ancestries.
        """
        return self._populate_images

    @populate_images.setter
    def populate_images(self, populate_images: bool) -> None:
        self.assert_mutable()
        self._populate_images = populate_images

    @override
    def load(self, dump: Dump) -> None:
        self.assert_mutable()
        assert_record(
            OptionalField(
                "populate_images",
                assert_bool() | assert_setattr(self, "populate_images"),
            )
        )(dump)

    @override
    def dump(self) -> DumpMapping[Dump]:
        return {
            "populate_images": self.populate_images,
        }



"""
Multiprocessing functionality.
"""

from concurrent import futures
from multiprocessing import get_context
from signal import signal, SIGINT, SIG_IGN


class ProcessPoolExecutor(futures.ProcessPoolExecutor):
    """
    Like :py:class:`concurrent.futures.ProcessPoolExecutor`, but with error handling and low memory consumption.

    This
    - uses the ``spawn`` method to create new processes, which is the Python 3.14 default.
    - ignores SIGINT/:py:class:`KeyboardInterrupt` delivered to it by the parent process to prevent unhelpful
      additional :py:class:`KeyboardInterrupt` being raised from within the process pool.
    """

    def __init__(
        self, max_workers: int | None = None, *, max_tasks_per_child: int | None = None
    ):
        super().__init__(
            initializer=signal,
            initargs=(SIGINT, SIG_IGN),
            max_workers=max_workers,
            max_tasks_per_child=max_tasks_per_child,
            mp_context=get_context("spawn"),
        )



"""
Test utilities for :py:mod:`betty.plugin`.
"""

from collections.abc import Sequence
from typing import Generic, TypeVar, Any

from typing_extensions import override

from betty.locale.localizable import Localizable, plain
from betty.locale.localizer import DEFAULT_LOCALIZER
from betty.machine_name import assert_machine_name, MachineName
from betty.plugin import Plugin
from betty.string import camel_case_to_kebab_case

_PluginT = TypeVar("_PluginT", bound=Plugin)


def assert_plugin_identifier(value: Any, plugin_type: type[_PluginT]) -> None:
    """
    Assert that something is a plugin identifier.
    """
    if isinstance(value, str):
        assert_machine_name()(value)
    else:
        assert issubclass(value, plugin_type)


class PluginTestBase(Generic[_PluginT]):
    """
    A base class for testing :py:class:`betty.plugin.Plugin` implementations.
    """

    def get_sut_class(self) -> type[_PluginT]:
        """
        Produce the class of the plugin under test.
        """
        raise NotImplementedError

    async def test_class_is_public(self) -> None:
        """
        Tests that the plugin class is public.
        """
        assert not self.get_sut_class().__name__.startswith("_"), (
            f"Failed asserting that plugin class {self.get_sut_class()} is public (its name must not start with an underscore)"
        )

    async def test_plugin_id(self) -> None:
        """
        Tests :py:meth:`betty.plugin.Plugin.plugin_id` implementations.
        """
        assert_machine_name()(self.get_sut_class().plugin_id())

    async def test_plugin_label(self) -> None:
        """
        Tests :py:meth:`betty.plugin.Plugin.plugin_label` implementations.
        """
        assert self.get_sut_class().plugin_label().localize(DEFAULT_LOCALIZER)

    async def test_plugin_description(self) -> None:
        """
        Tests :py:meth:`betty.plugin.Plugin.plugin_description` implementations.
        """
        description = self.get_sut_class().plugin_description()
        if description is not None:
            assert description.localize(DEFAULT_LOCALIZER)


class PluginInstanceTestBase(Generic[_PluginT], PluginTestBase[_PluginT]):
    """
    A base class for testing :py:class:`betty.plugin.Plugin` implementation instances.
    """

    def get_sut_instances(self) -> Sequence[_PluginT]:
        """
        Produce instances of the plugin under test.
        """
        raise NotImplementedError


class DummyPlugin(Plugin):
    """
    A dummy plugin implementation.
    """

    @override
    @classmethod
    def plugin_id(cls) -> MachineName:
        return camel_case_to_kebab_case(cls.__name__.strip("_"))

    @override
    @classmethod
    def plugin_label(cls) -> Localizable:
        return plain(cls.__name__)



"""
The Plugin API.

Plugins allow third-party code (e.g. your own Python package) to add functionality
to Betty.

Read more at :doc:`/development/plugin`.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TypeVar,
    Generic,
    Self,
    overload,
    TYPE_CHECKING,
    TypeAlias,
    Iterable,
)

from typing_extensions import override

from betty.error import UserFacingError
from betty.factory import TargetFactory, Factory, new
from betty.json.schema import Enum, Schema
from betty.locale.localizable import _, join, do_you_mean
from betty.machine_name import MachineName

if TYPE_CHECKING:
    from graphlib import TopologicalSorter
    from betty.locale.localizable import Localizable
    from collections.abc import AsyncIterator, Sequence, Mapping, Iterable, Iterator

_T = TypeVar("_T")


class PluginError(Exception):
    """
    Any error originating from the Plugin API.
    """

    pass


class Plugin(ABC):
    """
    A plugin.

    Plugins are identified by their :py:meth:`IDs <betty.plugin.Plugin.plugin_id>` as well as their types.
    Each must be able to uniquely identify the plugin within a plugin repository.

    To test your own subclasses, use :py:class:`betty.test_utils.plugin.PluginTestBase`.
    """

    @classmethod
    @abstractmethod
    def plugin_id(cls) -> MachineName:
        """
        Get the plugin ID.

        IDs are unique per plugin type:

        - A plugin repository **MUST** at most have a single plugin for any ID.
        - Different plugin repositories **MAY** each have a plugin with the same ID.
        """
        pass

    @classmethod
    @abstractmethod
    def plugin_label(cls) -> Localizable:
        """
        Get the human-readable short plugin label.
        """
        pass

    @classmethod
    def plugin_description(cls) -> Localizable | None:
        """
        Get the human-readable long plugin description.
        """
        return None


_PluginT = TypeVar("_PluginT", bound=Plugin)
_PluginCoT = TypeVar("_PluginCoT", bound=Plugin, covariant=True)


class OrderedPlugin(Generic[_PluginT], Plugin):
    """
    A plugin that can declare its order with respect to other plugins.
    """

    @classmethod
    def comes_before(cls) -> set[PluginIdentifier[_PluginT]]:
        """
        Get the plugins that this plugin comes before.

        The returned plugins come after this plugin.
        """
        return set()

    @classmethod
    def comes_after(cls) -> set[PluginIdentifier[_PluginT]]:
        """
        Get the plugins that this plugin comes after.

        The returned plugins come before this plugin.
        """
        return set()


class DependentPlugin(Generic[_PluginT], Plugin):
    """
    A plugin that can declare its dependency on other plugins.
    """

    @classmethod
    def depends_on(cls) -> set[PluginIdentifier[_PluginT]]:
        """
        The plugins this one depends on.
        """
        return set()


class ShorthandPluginBase(Plugin):
    """
    Allow shorthand declaration of plugins.
    """

    _plugin_id: MachineName
    _plugin_label: Localizable
    _plugin_description: Localizable | None = None

    @override
    @classmethod
    def plugin_id(cls) -> MachineName:
        return cls._plugin_id

    @override
    @classmethod
    def plugin_label(cls) -> Localizable:
        return cls._plugin_label

    @override
    @classmethod
    def plugin_description(cls) -> Localizable | None:
        return cls._plugin_description


PluginIdentifier: TypeAlias = type[_PluginT] | MachineName


def resolve_identifier(plugin_identifier: PluginIdentifier[Plugin]) -> MachineName:
    """
    Resolve a plugin identifier to a plugin ID.
    """
    if isinstance(plugin_identifier, str):
        return plugin_identifier
    return plugin_identifier.plugin_id()


class PluginNotFound(PluginError, UserFacingError):
    """
    Raised when a plugin cannot be found.
    """

    @classmethod
    def new(
        cls, plugin_id: MachineName, available_plugins: Sequence[type[Plugin]]
    ) -> Self:
        """
        Create a new instance.
        """
        return cls(
            join(
                _('Could not find a plugin "{plugin_id}".').format(plugin_id=plugin_id),
                do_you_mean(
                    *[f'"{plugin.plugin_id()}"' for plugin in available_plugins]
                ),
            )
        )


_PluginMixinOneT = TypeVar("_PluginMixinOneT")
_PluginMixinTwoT = TypeVar("_PluginMixinTwoT")
_PluginMixinThreeT = TypeVar("_PluginMixinThreeT")


class PluginIdToTypeMapping(Generic[_PluginCoT]):
    """
    Map plugin IDs to their types.
    """

    def __init__(self, id_to_type_mapping: Mapping[MachineName, type[_PluginCoT]]):
        self._id_to_type_mapping = id_to_type_mapping

    @classmethod
    async def new(cls, plugins: PluginRepository[_PluginCoT]) -> Self:
        """
        Create a new instance.
        """
        return cls({plugin.plugin_id(): plugin async for plugin in plugins})

    def get(
        self, plugin_identifier: MachineName | type[_PluginCoT]
    ) -> type[_PluginCoT]:
        """
        Get the type for the given plugin identifier.
        """
        if isinstance(plugin_identifier, type):
            return plugin_identifier
        try:
            return self._id_to_type_mapping[plugin_identifier]
        except KeyError:
            raise PluginNotFound.new(
                plugin_identifier, list(self._id_to_type_mapping.values())
            ) from None

    def __getitem__(
        self, plugin_identifier: MachineName | type[_PluginCoT]
    ) -> type[_PluginCoT]:
        return self.get(plugin_identifier)

    def __iter__(self) -> Iterator[MachineName]:
        yield from self._id_to_type_mapping


class PluginRepository(Generic[_PluginT], TargetFactory, ABC):
    """
    Discover and manage plugins.
    """

    def __init__(
        self, *, factory: Factory | None = None, schema_template: Schema | None = None
    ):
        self._factory = factory or new
        self._schema_template = schema_template
        self._plugin_id_schema: Enum | None = None

    async def resolve_identifier(
        self, plugin_identifier: PluginIdentifier[_PluginT]
    ) -> type[_PluginT]:
        """
        Resolve a plugin identifier to a plugin type.
        """
        if isinstance(plugin_identifier, type):
            return plugin_identifier
        return await self.get(plugin_identifier)

    async def resolve_identifiers(
        self, plugin_identifiers: Iterable[PluginIdentifier[_PluginT]]
    ) -> Sequence[type[_PluginT]]:
        """
        Resolve plugin identifiers to plugin types.
        """
        return [
            await self.resolve_identifier(plugin_identifier)
            for plugin_identifier in plugin_identifiers
        ]

    async def mapping(self) -> PluginIdToTypeMapping[_PluginT]:
        """
        Get the plugin ID to type mapping.
        """
        return await PluginIdToTypeMapping.new(self)

    @abstractmethod
    async def get(self, plugin_id: MachineName) -> type[_PluginT]:
        """
        Get a single plugin by its ID.

        :raises PluginNotFound: if no plugin can be found for the given ID.
        """
        pass

    @overload
    async def select(self) -> Sequence[type[_PluginT]]:
        pass

    @overload
    async def select(
        self, mixin_one: type[_PluginMixinOneT]
    ) -> Sequence[type[_PluginT & _PluginMixinOneT]]:
        pass

    @overload
    async def select(
        self, mixin_one: type[_PluginMixinOneT], mixin_two: type[_PluginMixinTwoT]
    ) -> Sequence[type[_PluginT & _PluginMixinOneT & _PluginMixinTwoT]]:
        pass

    @overload
    async def select(
        self,
        mixin_one: type[_PluginMixinOneT],
        mixin_two: type[_PluginMixinTwoT],
        mixin_three: type[_PluginMixinThreeT],
    ) -> Sequence[
        type[_PluginT & _PluginMixinOneT & _PluginMixinTwoT & _PluginMixinThreeT]
    ]:
        pass

    async def select(
        self,
        mixin_one: type[_PluginMixinOneT] | None = None,
        mixin_two: type[_PluginMixinTwoT] | None = None,
        mixin_three: type[_PluginMixinThreeT] | None = None,
    ) -> (
        Sequence[type[_PluginT]]
        | Sequence[type[_PluginT & _PluginMixinOneT]]
        | Sequence[type[_PluginT & _PluginMixinOneT & _PluginMixinTwoT]]
        | Sequence[
            type[_PluginT & _PluginMixinOneT & _PluginMixinTwoT & _PluginMixinThreeT]
        ]
    ):
        """
        Select a subset of the known plugins.

        When called without arguments, this returns all known plugins.
        When called with one or more arguments, this returns all known plugins
        that are also subclasses or **all** of the given mixins.

        This method is overloaded to provide for the majority use case of at most
        three mixins, because when using ``*args: *Ts``, we cannot unpack ``Ts`` into an
        :py:class:`basedtyping.Intersection` return type.
        """
        return [
            plugin
            async for plugin in self
            if self._select_plugin(plugin, mixin_one, mixin_two, mixin_three)
        ]

    def _select_plugin(
        self,
        plugin: type[_PluginT],
        mixin_one: type[_PluginMixinOneT] | None = None,
        mixin_two: type[_PluginMixinTwoT] | None = None,
        mixin_three: type[_PluginMixinThreeT] | None = None,
    ) -> bool:
        for mixin in (mixin_one, mixin_two, mixin_three):
            if mixin is None:
                continue
            if not issubclass(plugin, mixin):
                return False
        return True

    @abstractmethod
    def __aiter__(self) -> AsyncIterator[type[_PluginT]]:
        pass

    @overload
    async def new_target(self, cls: type[_T]) -> _T:
        pass

    @overload
    async def new_target(self, cls: MachineName) -> _PluginT:
        pass

    @override
    async def new_target(self, cls: type[_T] | MachineName) -> _T | _PluginT:
        return await self._factory(await self.get(cls) if isinstance(cls, str) else cls)

    @property
    async def plugin_id_schema(self) -> Enum:
        """
        Get the JSON schema for the IDs of the plugins in this repository.
        """
        if self._schema_template is None:
            raise RuntimeError(
                f"Cannot provide a JSON schema for {self} which has no schema template."
            )
        if self._plugin_id_schema is None:
            self._plugin_id_schema = Enum(
                *[plugin.plugin_id() async for plugin in self],  # noqa A002
                def_name=self._schema_template.def_name,
                title=self._schema_template.title,
                description=f"A {self._schema_template.title} plugin ID",
            )
        return self._plugin_id_schema


class CyclicDependencyError(PluginError):
    """
    Raised when plugins define a cyclic dependency, e.g. two plugins depend on each other.
    """

    def __init__(self, plugin_types: Iterable[type[Plugin]]):
        plugin_names = ", ".join([plugin.plugin_id() for plugin in plugin_types])
        super().__init__(
            f"The following plugins have cyclic dependencies: {plugin_names}"
        )


async def sort_ordered_plugin_graph(
    sorter: TopologicalSorter[type[_PluginT]],
    plugins: PluginRepository[_PluginT & OrderedPlugin[_PluginT]],
    entry_point_plugins: Iterable[type[_PluginT & OrderedPlugin[_PluginT]]],
) -> None:
    """
    Build a graph of the given plugins.
    """
    entry_point_plugins = list(entry_point_plugins)
    for entry_point_plugin in entry_point_plugins:
        sorter.add(entry_point_plugin)
        for before_identifier in entry_point_plugin.comes_before():
            before = (
                await plugins.get(before_identifier)
                if isinstance(before_identifier, str)
                else before_identifier
            )
            if before in entry_point_plugins:
                sorter.add(before, entry_point_plugin)
        for after_identifier in entry_point_plugin.comes_after():
            after = (
                await plugins.get(after_identifier)
                if isinstance(after_identifier, str)
                else after_identifier
            )
            if after in entry_point_plugins:
                sorter.add(entry_point_plugin, after)


async def sort_dependent_plugin_graph(
    sorter: TopologicalSorter[type[_PluginT]],
    plugins: PluginRepository[_PluginT],
    entry_point_plugins: Iterable[type[_PluginT & DependentPlugin[_PluginT]]],
) -> Sequence[type[_PluginT & DependentPlugin[_PluginT]]]:
    """
    Build a graph of the given plugins and their dependencies.
    """
    updated_entry_point_plugins = []
    for entry_point_plugin in entry_point_plugins:
        if entry_point_plugin not in updated_entry_point_plugins:
            updated_entry_point_plugins.append(entry_point_plugin)
        dependencies = await plugins.resolve_identifiers(
            entry_point_plugin.depends_on()
        )
        for dependency in dependencies:
            if dependency not in updated_entry_point_plugins:
                updated_entry_point_plugins.append(
                    dependency,  # type: ignore[arg-type]
                )
        sorter.add(entry_point_plugin, *dependencies)
        await sort_dependent_plugin_graph(
            sorter,
            plugins,
            # We have not quite figured out how to type this correctly, so ignore any errors for now.
            dependencies,  # type: ignore[arg-type]
        )
    return updated_entry_point_plugins



"""
Provide caching that persists cache items to files.
"""

from __future__ import annotations

import asyncio
import shutil
from abc import abstractmethod
from contextlib import suppress
from functools import partial
from os import utime
from pickle import dumps, loads
from typing import (
    Generic,
    Self,
    TYPE_CHECKING,
    TypeVar,
    final,
    AsyncContextManager,
    overload,
    Literal,
)

import aiofiles
from aiofiles.ospath import getmtime
from typing_extensions import override

from betty.cache import CacheItem, CacheItemValueSetter
from betty.cache._base import _CommonCacheBase, _CommonCacheBaseState
from betty.hashid import hashid
from betty.typing import processsafe

if TYPE_CHECKING:
    from multiprocessing.managers import SyncManager
    from pathlib import Path
    from collections.abc import Sequence

_CacheItemValueCoT = TypeVar("_CacheItemValueCoT", covariant=True)
_CacheItemValueContraT = TypeVar("_CacheItemValueContraT", contravariant=True)


class _FileCacheItem(CacheItem[_CacheItemValueCoT], Generic[_CacheItemValueCoT]):
    __slots__ = "_modified", "_path"

    def __init__(
        self,
        modified: int | float,
        path: Path,
    ):
        self._modified = modified
        self._path = path

    @override
    @property
    def modified(self) -> int | float:
        return self._modified

    @override
    async def value(self) -> _CacheItemValueCoT:
        async with aiofiles.open(self._path, "rb") as f:
            value_bytes = await f.read()
        return await self._load_value(value_bytes)

    @abstractmethod
    async def _load_value(self, value_bytes: bytes) -> _CacheItemValueCoT:
        pass


@final
class _PickledFileCacheItem(
    _FileCacheItem[_CacheItemValueCoT], Generic[_CacheItemValueCoT]
):
    @override
    async def _load_value(self, value_bytes: bytes) -> _CacheItemValueCoT:
        return loads(value_bytes)  # type: ignore[no-any-return]


@final
class _BinaryFileCacheItem(_FileCacheItem[bytes]):
    @override
    async def _load_value(self, value_bytes: bytes) -> bytes:
        return value_bytes


class _FileCache(
    _CommonCacheBase[_CacheItemValueContraT], Generic[_CacheItemValueContraT]
):
    """
    Provide a cache that persists cache items on a file system.
    """

    _cache_item_cls: type[_FileCacheItem[_CacheItemValueContraT]]

    def __init__(
        self,
        cache_directory_path: Path,
        *,
        scopes: Sequence[str] | None = None,
        manager: SyncManager | _CommonCacheBaseState[Self] | None = None,
    ):
        super().__init__(scopes=scopes, manager=manager)
        self._root_path = cache_directory_path

    @override
    def with_scope(self, scope: str) -> Self:
        return type(self)(
            self._root_path,
            scopes=(*self._scopes, scope),
            manager=_CommonCacheBaseState(
                self._cache_lock, self._cache_item_lock_ledger
            ),
        )

    def _cache_item_file_path(
        self, cache_item_id: str, suffix: str | None = None
    ) -> Path:
        cache_item_file_path = self._path / hashid(cache_item_id)
        if suffix is not None:
            assert suffix.startswith(".")
            cache_item_file_path = cache_item_file_path.parent / (
                cache_item_file_path.name + suffix
            )
        return cache_item_file_path

    @abstractmethod
    def _dump_value(self, value: _CacheItemValueContraT) -> bytes:
        pass

    @override
    async def _get(
        self, cache_item_id: str, *, suffix: str | None = None
    ) -> CacheItem[_CacheItemValueContraT] | None:
        try:
            cache_item_file_path = self._cache_item_file_path(cache_item_id, suffix)
            return self._cache_item_cls(
                await getmtime(cache_item_file_path),
                cache_item_file_path,
            )
        except OSError:
            return None

    @override
    async def _set(
        self,
        cache_item_id: str,
        value: _CacheItemValueContraT,
        *,
        suffix: str | None = None,
        modified: int | float | None = None,
    ) -> None:
        value = self._dump_value(value)
        cache_item_file_path = self._cache_item_file_path(cache_item_id, suffix)
        try:
            await self._write(cache_item_file_path, value, modified)
        except FileNotFoundError:
            await aiofiles.os.makedirs(cache_item_file_path.parent, exist_ok=True)
            await self._write(cache_item_file_path, value, modified)

    async def _write(
        self,
        cache_item_file_path: Path,
        value: bytes,
        modified: int | float | None = None,
    ) -> None:
        async with aiofiles.open(cache_item_file_path, "wb") as f:
            await f.write(value)
        if modified is not None:
            await asyncio.to_thread(utime, cache_item_file_path, (modified, modified))

    @override
    async def _delete(self, cache_item_id: str, *, suffix: str | None = None) -> None:
        with suppress(FileNotFoundError):
            await aiofiles.os.remove(self._cache_item_file_path(cache_item_id, suffix))

    @override
    async def _clear(self) -> None:
        with suppress(FileNotFoundError):
            await asyncio.to_thread(shutil.rmtree, self._path)

    @property
    def _path(self) -> Path:
        return self._root_path.joinpath(*self._scopes)

    @overload
    def getset(
        self, cache_item_id: str, *, suffix: str | None = None
    ) -> AsyncContextManager[
        tuple[
            CacheItem[_CacheItemValueContraT] | None,
            CacheItemValueSetter[_CacheItemValueContraT],
        ]
    ]:
        pass

    @overload
    def getset(
        self,
        cache_item_id: str,
        *,
        suffix: str | None = None,
        wait: Literal[False] = False,
    ) -> AsyncContextManager[
        tuple[
            CacheItem[_CacheItemValueContraT] | None,
            CacheItemValueSetter[_CacheItemValueContraT] | None,
        ]
    ]:
        pass

    @override
    def getset(
        self, cache_item_id: str, *, suffix: str | None = None, wait: bool = True
    ) -> AsyncContextManager[
        tuple[
            CacheItem[_CacheItemValueContraT] | None,
            CacheItemValueSetter[_CacheItemValueContraT] | None,
        ]
    ]:
        return self._getset(
            cache_item_id,
            partial(self._get, suffix=suffix),
            partial(self._set, suffix=suffix),
            wait=wait,
        )


@final
@processsafe
class PickledFileCache(
    _FileCache[_CacheItemValueContraT], Generic[_CacheItemValueContraT]
):
    """
    Provide a cache that pickles values and persists them to files.
    """

    _cache_item_cls = _PickledFileCacheItem

    @override
    def _dump_value(self, value: _CacheItemValueContraT) -> bytes:
        return dumps(value)


@final
@processsafe
class BinaryFileCache(_FileCache[bytes]):
    """
    Provide a cache that persists bytes values to binary files.
    """

    _cache_item_cls = _BinaryFileCacheItem

    @override
    def _dump_value(self, value: bytes) -> bytes:
        return value

    @property
    def path(self) -> Path:
        """
        The path to the cache's root directory.
        """
        return self._path

    def cache_item_file_path(
        self, cache_item_id: str, suffix: str | None = None
    ) -> Path:
        """
        Get the file path for a cache item with the given ID.

        The cache item itself may or may not exist.
        """
        return self._cache_item_file_path(cache_item_id, suffix)



"""
Test utilities for :py:mod:`betty.ancestry.locale`.
"""

from betty.ancestry.locale import HasLocale


class DummyHasLocale(HasLocale):
    """
    A dummy :py:class:`betty.ancestry.locale.HasLocale` implementation.
    """

    pass



"""
Data types representing files on disk.
"""

from __future__ import annotations

from typing import final, Iterable, MutableSequence, TYPE_CHECKING

from typing_extensions import override

from betty.ancestry.description import HasDescription
from betty.ancestry.has_citations import HasCitations
from betty.ancestry.has_notes import HasNotes
from betty.ancestry.link import HasLinks, Link
from betty.ancestry.media_type import HasMediaType
from betty.locale.localizable import (
    _,
    ShorthandStaticTranslations,
    Localizable,
    ngettext,
)
from betty.model import UserFacingEntity, Entity
from betty.model.association import (
    BidirectionalToMany,
    ToManyAssociates,
)
from betty.plugin import ShorthandPluginBase
from betty.privacy import HasPrivacy, Privacy

if TYPE_CHECKING:
    from betty.mutability import Mutable
    from betty.json.linked_data import JsonLdObject
    from betty.project import Project
    from betty.serde.dump import DumpMapping, Dump
    from betty.copyright_notice import CopyrightNotice
    from betty.license import License
    from betty.ancestry.citation import Citation
    from betty.ancestry.note import Note
    from betty.ancestry.file_reference import FileReference  # noqa F401
    from betty.media_type import MediaType
    from pathlib import Path


@final
class File(
    ShorthandPluginBase,
    HasDescription,
    HasPrivacy,
    HasLinks,
    HasMediaType,
    HasNotes,
    HasCitations,
    UserFacingEntity,
    Entity,
):
    """
    A file on disk.

    This includes but is not limited to:

    - images
    - video
    - audio
    - PDF documents
    """

    _plugin_id = "file"
    _plugin_label = _("File")

    referees = BidirectionalToMany["File", "FileReference"](
        "betty.ancestry.file:File",
        "referees",
        "betty.ancestry.file_reference:FileReference",
        "file",
        title="Referees",
        description="The entities referencing this file",
        linked_data_embedded=True,
    )

    #: The copyright notice for this file.
    copyright_notice: CopyrightNotice | None

    #: The license for this file.
    license: License | None

    def __init__(
        self,
        path: Path,
        *,
        id: str | None = None,  # noqa A002  # noqa A002
        name: str | None = None,
        media_type: MediaType | None = None,
        description: ShorthandStaticTranslations | None = None,
        notes: ToManyAssociates[Note] | None = None,
        citations: ToManyAssociates[Citation] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        links: MutableSequence[Link] | None = None,
        copyright_notice: CopyrightNotice | None = None,
        license: License | None = None,  # noqa A002
    ):
        super().__init__(
            id,
            media_type=media_type,
            description=description,
            notes=notes,
            citations=citations,
            privacy=privacy,
            public=public,
            private=private,
            links=links,
        )
        self._path = path
        self._name = name
        self.copyright_notice = copyright_notice
        self.license = license

    @override
    def get_mutable_instances(self) -> Iterable[Mutable]:
        if self.copyright_notice is not None:
            yield self.copyright_notice
        if self.license is not None:
            yield self.license

    @property
    def name(self) -> str:
        """
        The file name.
        """
        return self._name or self.path.name

    @override
    @classmethod
    def plugin_label_plural(cls) -> Localizable:
        return _("Files")

    @override
    @classmethod
    def plugin_label_count(cls, count: int) -> Localizable:
        return ngettext("{count} file", "{count} files", count).format(count=str(count))

    @property
    def path(self) -> Path:
        """
        The file's path on disk.
        """
        return self._path

    @override
    @property
    def label(self) -> Localizable:
        return self.description or super().label

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "copyrightNotice",
            await project.copyright_notice_repository.plugin_id_schema,
            False,
        )
        schema.add_property(
            "license", await (await project.license_repository).plugin_id_schema, False
        )
        return schema

    @override
    async def dump_linked_data(self, project: Project) -> DumpMapping[Dump]:
        dump = await super().dump_linked_data(project)
        if self.copyright_notice:
            dump["copyrightNotice"] = self.copyright_notice.plugin_id()
        if self.license:
            dump["license"] = self.license.plugin_id()
        return dump



"""
Provide concrete gender implementations.
"""

from betty.ancestry.gender import Gender
from betty.locale.localizable import _
from betty.plugin import ShorthandPluginBase


class Female(ShorthandPluginBase, Gender):
    """
    A female person.
    """

    _plugin_id = "female"
    _plugin_label = _("Female")


class Male(ShorthandPluginBase, Gender):
    """
    A male person.
    """

    _plugin_id = "male"
    _plugin_label = _("Male")


class NonBinary(ShorthandPluginBase, Gender):
    """
    A non-binary person.
    """

    _plugin_id = "non-binary"
    _plugin_label = _("Non-binary")


class Unknown(ShorthandPluginBase, Gender):
    """
    A person of an unknown gender.
    """

    _plugin_id = "unknown"
    _plugin_label = _("Unknown")



"""Define Betty's core application functionality."""

from __future__ import annotations

import multiprocessing
from contextlib import asynccontextmanager
from os import environ
from pathlib import Path
from typing import TYPE_CHECKING, Self, Any, final, TypeVar, cast

import aiohttp
from aiofiles.tempfile import TemporaryDirectory
from typing_extensions import override

from betty import fs
from betty.app import config
from betty.app.config import AppConfiguration
from betty.app.factory import AppDependentFactory
from betty.assets import AssetRepository
from betty.cache.file import BinaryFileCache, PickledFileCache
from betty.cache.no_op import NoOpCache
from betty.config import Configurable, assert_configuration_file
from betty.factory import new, TargetFactory
from betty.fetch import Fetcher, http
from betty.fetch.static import StaticFetcher
from betty.fs import HOME_DIRECTORY_PATH
from betty.license import License, LICENSE_REPOSITORY
from betty.license.licenses import SpdxLicenseRepository
from betty.locale import DEFAULT_LOCALE
from betty.locale.localizer import Localizer, LocalizerRepository
from betty.multiprocessing import ProcessPoolExecutor
from betty.plugin.proxy import ProxyPluginRepository
from betty.service import ServiceProvider, service, ServiceFactory, StaticService
from betty.typing import processsafe

if TYPE_CHECKING:
    from concurrent import futures
    from multiprocessing.managers import SyncManager
    from betty.plugin import PluginRepository
    from betty.cache import Cache
    from collections.abc import AsyncIterator

_T = TypeVar("_T")


@final
@processsafe
class App(Configurable[AppConfiguration], TargetFactory, ServiceProvider):
    """
    The Betty application.
    """

    def __init__(
        self,
        configuration: AppConfiguration,
        cache_directory_path: Path,
        *,
        cache_factory: ServiceFactory[Self, Cache[Any]],
        fetcher: Fetcher | None = None,
        process_pool: futures.ProcessPoolExecutor | None = None,
        multiprocessing_manager: SyncManager | None = None,
    ):
        cls = type(self)
        super().__init__(configuration=configuration)
        if fetcher is not None:
            cls.fetcher.override(self, fetcher)
        if process_pool is not None:
            cls.process_pool.override(self, process_pool)
        if multiprocessing_manager is not None:
            cls.multiprocessing_manager.override(self, multiprocessing_manager)
        self._cache_directory_path = cache_directory_path
        cls.cache.override_factory(self, cache_factory)

    def __getstate__(self) -> dict[str, Any]:
        cls = type(self)
        return {
            **super().__getstate__(),
            "_bootstrapped": True,
            "_cache_directory_path": self._cache_directory_path,
            "_configuration": self._configuration,
            **cls.binary_file_cache.get_state(self),
            **cls.cache.get_state(self),
        }

    @classmethod
    @asynccontextmanager
    async def new_from_environment(cls) -> AsyncIterator[Self]:
        """
        Create a new application from the environment.
        """
        configuration = AppConfiguration()
        if config.CONFIGURATION_FILE_PATH.exists():
            (await assert_configuration_file(configuration))(
                config.CONFIGURATION_FILE_PATH
            )
        yield cls(
            configuration,
            Path(environ.get("BETTY_CACHE_DIRECTORY", HOME_DIRECTORY_PATH / "cache")),
            cache_factory=lambda app: PickledFileCache[Any](
                app._cache_directory_path, manager=app.multiprocessing_manager
            ),
        )

    @classmethod
    @asynccontextmanager
    async def new_temporary(
        cls,
        *,
        cache_factory: ServiceFactory[Self, Cache[Any]] | None = None,
        fetcher: Fetcher | None = None,
        process_pool: futures.ProcessPoolExecutor | None = None,
        multiprocessing_manager: SyncManager | None = None,
    ) -> AsyncIterator[Self]:
        """
        Create a new, temporary, isolated application.

        The application will not use any persistent caches, or leave
        any traces on the system.
        """
        async with (
            TemporaryDirectory() as cache_directory_path_str,
        ):
            yield cls(
                AppConfiguration(),
                Path(cache_directory_path_str),
                cache_factory=cache_factory or StaticService(NoOpCache()),
                fetcher=fetcher or StaticFetcher(),
                process_pool=process_pool,
                multiprocessing_manager=multiprocessing_manager,
            )

    @service
    def assets(self) -> AssetRepository:
        """
        The assets file system.
        """
        return AssetRepository(fs.ASSETS_DIRECTORY_PATH)

    @service
    async def localizer(self) -> Localizer:
        """
        Get the application's localizer.
        """
        return await self.localizers.get_negotiated(
            self.configuration.locale or DEFAULT_LOCALE
        )

    @service
    def localizers(self) -> LocalizerRepository:
        """
        The available localizers.
        """
        return LocalizerRepository(self.assets)

    @service
    async def http_client(self) -> aiohttp.ClientSession:
        """
        The HTTP client.
        """
        http_client = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(limit_per_host=5),
            headers={
                "User-Agent": "Betty (https://betty.readthedocs.io/)",
            },
        )

        async def _shutdown(wait: bool) -> None:
            await http_client.close()

        self._shutdown_stack.append(_shutdown)
        return http_client

    @service
    async def fetcher(self) -> Fetcher:
        """
        The fetcher.
        """
        return http.HttpFetcher(
            await self.http_client,
            self.cache.with_scope("fetch"),
            self.binary_file_cache.with_scope("fetch"),
        )

    @service(shared=True)
    def cache(self) -> Cache[Any]:
        """
        The cache.
        """
        raise NotImplementedError

    @service(shared=True)
    def binary_file_cache(self) -> BinaryFileCache:
        """
        The binary file cache.
        """
        return BinaryFileCache(
            self._cache_directory_path, manager=self.multiprocessing_manager
        )

    @service
    def process_pool(self) -> futures.ProcessPoolExecutor:
        """
        The shared process pool.

        Use this to run CPU/computationally-heavy tasks in other processes.
        """
        process_pool = ProcessPoolExecutor()

        async def _shutdown(wait: bool) -> None:
            process_pool.shutdown(wait, cancel_futures=not wait)

        self._shutdown_stack.append(_shutdown)
        return process_pool

    @service
    def multiprocessing_manager(self) -> SyncManager:
        """
        The multiprocessing manager.

        Use this to create process-safe synchronization primitives.
        """
        manager = multiprocessing.Manager()

        async def _shutdown(wait: bool) -> None:
            manager.shutdown(wait)

        self._shutdown_stack.append(_shutdown)
        return manager

    @override
    async def new_target(self, cls: type[_T]) -> _T:
        """
        Create a new instance.

        :return:
            #. If ``cls`` extends :py:class:`betty.app.factory.AppDependentFactory`, this will call return ``cls``'s
                ``new()``'s return value.
            #. If ``cls`` extends :py:class:`betty.factory.IndependentFactory`, this will call return ``cls``'s
                ``new()``'s return value.
            #. Otherwise ``cls()`` will be called without arguments, and the resulting instance will be returned.

        :raises FactoryError: raised when ``cls`` could not be instantiated.
        """
        if issubclass(cls, AppDependentFactory):
            return cast("_T", await cls.new_for_app(self))
        return await new(cls)

    @service
    async def spdx_license_repository(self) -> PluginRepository[License]:
        """
        The SPDX licenses available to this application.
        """
        return ProxyPluginRepository(
            LICENSE_REPOSITORY,
            SpdxLicenseRepository(
                binary_file_cache=self.binary_file_cache.with_scope("spdx"),
                fetcher=await self.fetcher,
                localizer=await self.localizer,
                factory=self.new_target,
                process_pool=self.process_pool,
                manager=self.multiprocessing_manager,
            ),
        )



"""
Provide the Command Line Interface.
"""

from __future__ import annotations

import logging
from abc import abstractmethod
from contextlib import suppress
from functools import wraps
from pathlib import Path
from typing import Any, Concatenate, ParamSpec, TYPE_CHECKING, overload, TypeVar, cast

import asyncclick as click

from betty import about
from betty.assertion import assert_path, assert_none, assert_or
from betty.assertion.error import AssertionFailed
from betty.cli.error import user_facing_error_to_bad_parameter
from betty.config import assert_configuration_file
from betty.error import UserFacingError, FileNotFound
from betty.locale.localizable import _
from betty.plugin import Plugin, PluginRepository
from betty.plugin.entry_point import EntryPointPluginRepository
from betty.plugin.proxy import ProxyPluginRepository
from betty.project import Project
from betty.project.config import ProjectConfiguration
from betty.serde.format import FORMAT_REPOSITORY

if TYPE_CHECKING:
    from betty.cli import ContextAppObject
    from collections.abc import Callable, Coroutine

_T = TypeVar("_T")
_P = ParamSpec("_P")


class Command(Plugin):
    """
    Define a CLI command plugin.

    Read more about :doc:`/development/plugin/command`.
    """

    @abstractmethod
    async def click_command(self) -> click.Command:
        """
        Get the plugin's Click command.
        """
        pass


COMMAND_REPOSITORY: PluginRepository[Command] = ProxyPluginRepository(
    EntryPointPluginRepository("betty.command"),
    *(
        [EntryPointPluginRepository("betty.dev.command")]
        if about.is_development()
        else []
    ),
)
"""
The Command Line Interface command repository.

Read more about :doc:`/development/plugin/command`.
"""


def _command_build_init_ctx_verbosity(
    betty_logger_level: int,
    root_logger_level: int | None = None,
) -> Callable[[click.Context, click.Option | click.Parameter | None, bool], None]:
    def _init_ctx_verbosity(
        _: click.Context,
        __: click.Option | click.Parameter | None = None,
        is_verbose: bool = False,
    ) -> None:
        if is_verbose:
            for logger_name, logger_level in (
                ("betty", betty_logger_level),
                (None, root_logger_level),
            ):
                logger = logging.getLogger(logger_name)
                if (
                    logger_level is not None
                    and logger.getEffectiveLevel() > logger_level
                ):
                    logger.setLevel(logger_level)

    return _init_ctx_verbosity


@overload
def command(
    name: Callable[..., Coroutine[Any, Any, Any]],
) -> click.Command:
    pass


@overload
def command(
    name: str | None,
    **attrs: Any,
) -> Callable[[Callable[..., Coroutine[Any, Any, Any]]], click.Command]:
    pass


@overload
def command(
    name: None = None,
    **attrs: Any,
) -> Callable[[Callable[..., Coroutine[Any, Any, Any]]], click.Command]:
    pass


@overload
def command(
    name: str | None = None,
    **attrs: Any,
) -> Callable[[Callable[..., Coroutine[Any, Any, Any]]], click.Command]:
    pass


def command(
    name: str | None | Callable[..., Coroutine[Any, Any, Any]] = None,
    **attrs: Any,
) -> click.Command | Callable[[Callable[..., Coroutine[Any, Any, Any]]], click.Command]:
    """
    Mark something a Betty command.

    This is almost identical to :py:func:`asyncclick.command`.

    Functions decorated with ``@command`` may choose to raise :py:class:`betty.error.UserFacingError`, which will
    automatically be localized and reraised as :py:class:`asyncclick.ClickException`.

    Read more about :doc:`/development/plugin/command`.
    """

    def decorator(f: Callable[_P, Coroutine[Any, Any, Any]]) -> click.Command:
        @click.command(cast("str | None", name), **attrs)
        @click.option(
            "-v",
            "--verbose",
            is_eager=True,
            default=False,
            is_flag=True,
            expose_value=False,
            help="Show verbose output, including informative log messages.",
            callback=_command_build_init_ctx_verbosity(logging.INFO),
        )
        @click.option(
            "-vv",
            "--more-verbose",
            "more_verbose",
            is_eager=True,
            default=False,
            is_flag=True,
            expose_value=False,
            help="Show more verbose output, including debug log messages.",
            callback=_command_build_init_ctx_verbosity(logging.DEBUG),
        )
        @click.option(
            "-vvv",
            "--most-verbose",
            "most_verbose",
            is_eager=True,
            default=False,
            is_flag=True,
            expose_value=False,
            help="Show most verbose output, including all log messages.",
            callback=_command_build_init_ctx_verbosity(logging.NOTSET, logging.NOTSET),
        )
        @click.pass_obj
        @wraps(f)
        async def _command(
            obj: ContextAppObject, *args: _P.args, **kwargs: _P.kwargs
        ) -> Any:
            try:
                return await f(*args, **kwargs)
            except UserFacingError as error:
                raise click.ClickException(error.localize(obj.localizer)) from error

        return cast("click.Command", _command)

    if callable(name):
        return decorator(name)
    return decorator


_ReturnT = TypeVar("_ReturnT")


def parameter_callback(
    f: Callable[Concatenate[_T, _P], _ReturnT], *args: _P.args, **kwargs: _P.kwargs
) -> Callable[[click.Context, click.Parameter, _T], _ReturnT]:
    """
    Convert a callback that takes a parameter (option, argument) value and returns it after processing.

    This handles errors so Click can gracefully exit.
    """
    from betty.cli import ctx_app_object

    def _callback(ctx: click.Context, __: click.Parameter, value: _T) -> _ReturnT:
        with user_facing_error_to_bad_parameter(ctx_app_object(ctx).localizer):
            return f(value, *args, **kwargs)

    return _callback


async def _read_project_configuration(
    project: Project, provided_configuration_file_path_str: Path | None
) -> None:
    project_directory_path = Path.cwd()
    if provided_configuration_file_path_str is None:
        try_configuration_file_paths = [
            project_directory_path / f"betty{extension}"
            for extension in await FORMAT_REPOSITORY.extensions()
        ]
        for try_configuration_file_path in try_configuration_file_paths:
            with suppress(FileNotFound):
                return await _read_project_configuration_file(
                    project, try_configuration_file_path
                )
        raise AssertionFailed(
            _(
                "Could not find any of the following configuration files in {project_directory_path}: {configuration_file_names}."
            ).format(
                configuration_file_names=", ".join(
                    str(try_configuration_file_path.relative_to(project_directory_path))
                    for try_configuration_file_path in try_configuration_file_paths
                ),
                project_directory_path=str(project_directory_path),
            )
        )
    else:
        await _read_project_configuration_file(
            project,
            (project_directory_path / provided_configuration_file_path_str)
            .expanduser()
            .resolve(),
        )


async def _read_project_configuration_file(
    project: Project, configuration_file_path: Path
) -> None:
    localizer = await project.app.localizer
    logger = logging.getLogger(__name__)
    assert_configuration = await assert_configuration_file(project.configuration)
    try:
        assert_configuration(configuration_file_path)
    except UserFacingError as error:
        logger.debug(error.localize(localizer))
        raise
    else:
        project.configuration.configuration_file_path = configuration_file_path
        logger.info(
            localizer._(
                "Loaded the configuration from {configuration_file_path}."
            ).format(configuration_file_path=str(configuration_file_path)),
        )


def project_option(
    f: Callable[Concatenate[Project, _P], Coroutine[Any, Any, _ReturnT]],
) -> Callable[_P, Coroutine[Any, Any, _ReturnT]]:
    """
    Decorate a command that requires a :py:class:`betty.project.Project`.
    """

    @click.option(
        "--configuration",
        "-c",
        "configuration_file_path",
        help="The path to a Betty project configuration file. Defaults to betty.json|yaml|yml in the current working directory.",
        callback=parameter_callback(assert_or(assert_path(), assert_none())),
    )
    @click.pass_obj
    @wraps(f)
    async def _project_option(
        obj: ContextAppObject,
        *args: Any,
        configuration_file_path: Path | None,
        **kwargs: Any,
    ) -> _ReturnT:
        project = await Project.new(
            obj.app, configuration=await ProjectConfiguration.new(Path())
        )
        await _read_project_configuration(project, configuration_file_path)
        async with project:
            return await f(project, *args, **kwargs)

    return _project_option  # type: ignore[return-value]



"""
Provide Cotton Candy's search functionality.
"""

from __future__ import annotations

import json
from abc import ABC
from asyncio import gather
from dataclasses import dataclass
from inspect import getmembers
from typing import TYPE_CHECKING, TypeVar, Generic, final, cast

import aiofiles
from typing_extensions import override

from betty.ancestry.file import File
from betty.ancestry.has_notes import HasNotes
from betty.ancestry.person import Person
from betty.ancestry.place import Place
from betty.ancestry.source import Source
from betty.locale.localizable import StaticTranslationsLocalizable, Localizable
from betty.locale.localizable import StaticTranslationsLocalizableAttr
from betty.model import Entity
from betty.privacy import is_private
from betty.typing import internal

if TYPE_CHECKING:
    from betty.machine_name import MachineName
    from betty.project import Project
    from betty.jinja2 import Environment
    from betty.ancestry import Ancestry
    from betty.locale.localizer import Localizer
    from betty.job import Context
    from collections.abc import Iterable, Sequence

_EntityT = TypeVar("_EntityT", bound=Entity)


async def generate_search_index(
    project: Project,
    result_container_template: Localizable,
    results_container_template: Localizable,
    *,
    job_context: Context,
) -> None:
    await gather(
        *(
            _generate_search_index_for_locale(
                project,
                result_container_template,
                results_container_template,
                locale,
                job_context=job_context,
            )
            for locale in project.configuration.locales
        )
    )


async def _generate_search_index_for_locale(
    project: Project,
    result_container_template: Localizable,
    results_container_template: Localizable,
    locale: str,
    *,
    job_context: Context,
) -> None:
    localizers = await project.localizers
    localizer = await localizers.get(locale)
    search_index = {
        "resultContainerTemplate": result_container_template.localize(localizer),
        "resultsContainerTemplate": results_container_template.localize(localizer),
        "index": [
            {
                "entityTypeId": entry.entity_type_id,
                "text": " ".join(entry.text),
                "result": entry.result,
            }
            for entry in await Index(
                project.ancestry,
                await project.jinja2_environment,
                job_context,
                localizer,
            ).build()
        ],
    }
    search_index_json = json.dumps(search_index)
    async with aiofiles.open(
        project.configuration.localize_www_directory_path(locale) / "search-index.json",
        mode="w",
    ) as f:
        await f.write(search_index_json)


def _static_translations_to_text(
    translations: StaticTranslationsLocalizable,
) -> set[str]:
    return {
        word
        for translation in translations.translations.values()
        for word in translation.strip().lower().split()
    }


class _EntityTypeIndexer(Generic[_EntityT], ABC):
    def text(self, localizer: Localizer, entity: _EntityT) -> set[str]:
        text = {entity.id.lower()}

        # Each note is owner by a single other entity, so index it as part of that entity.
        if isinstance(entity, HasNotes):
            for note in entity.notes:
                text.update(_static_translations_to_text(note.text))

        for attr_name, class_attr_value in getmembers(type(entity)):
            if isinstance(class_attr_value, StaticTranslationsLocalizableAttr):
                text.update(
                    _static_translations_to_text(
                        cast(
                            "StaticTranslationsLocalizable", getattr(entity, attr_name)
                        )
                    )
                )

        return text


class _PersonIndexer(_EntityTypeIndexer[Person]):
    @override
    def text(self, localizer: Localizer, entity: Person) -> set[str]:
        text = super().text(localizer, entity)
        for name in entity.names:
            if name.individual is not None:
                text.update(set(name.individual.lower().split()))
            if name.affiliation is not None:
                text.update(set(name.affiliation.lower().split()))
        return text


class _PlaceIndexer(_EntityTypeIndexer[Place]):
    @override
    def text(self, localizer: Localizer, entity: Place) -> set[str]:
        text = super().text(localizer, entity)
        for name in entity.names:
            text.update(_static_translations_to_text(name.name))
        return text


class _FileIndexer(_EntityTypeIndexer[File]):
    @override
    def text(self, localizer: Localizer, entity: File) -> set[str]:
        text = super().text(localizer, entity)
        text.update(entity.path.name.strip().lower().split())
        if entity.description:
            text.update(entity.description.localize(localizer).strip().lower().split())
        return text


class _SourceIndexer(_EntityTypeIndexer[Source]):
    pass


@final
@dataclass(frozen=True)
class _Entry:
    entity_type_id: MachineName
    result: str
    text: set[str]


@internal
class Index:
    """
    Build search indexes.
    """

    def __init__(
        self,
        ancestry: Ancestry,
        jinja2_environment: Environment,
        job_context: Context | None,
        localizer: Localizer,
    ):
        self._ancestry = ancestry
        self._jinja2_environment = jinja2_environment
        self._job_context = job_context
        self._localizer = localizer

    async def build(self) -> Sequence[_Entry]:
        """
        Build the search index.
        """
        return [
            entry
            for entries in await gather(
                self._build_entities(_PersonIndexer(), Person),
                self._build_entities(_PlaceIndexer(), Place),
                self._build_entities(_FileIndexer(), File),
                self._build_entities(_SourceIndexer(), Source),
            )
            for entry in entries
            if entry is not None
        ]

    async def _build_entities(
        self, indexer: _EntityTypeIndexer[_EntityT], entity_type: type[_EntityT]
    ) -> Iterable[_Entry | None]:
        return await gather(
            *(
                self._build_entity(indexer, entity)
                for entity in self._ancestry[entity_type]
            )
        )

    async def _build_entity(
        self, indexer: _EntityTypeIndexer[_EntityT], entity: _EntityT
    ) -> _Entry | None:
        if is_private(entity):
            return None
        text = indexer.text(self._localizer, entity)
        if not text:
            return None
        return _Entry(entity.plugin_id(), await self._render_entity(entity), text)

    async def _render_entity(self, entity: Entity) -> str:
        return await self._jinja2_environment.select_template(
            [
                f"search/result--{entity.plugin_id()}.html.j2",
                "search/result.html.j2",
            ]
        ).render_async(
            {
                "job_context": self._job_context,
                "localizer": self._localizer,
                "entity": entity,
            }
        )



"""
Provide utilities for running jobs concurrently.
"""

from __future__ import annotations

from datetime import datetime
from typing import Any, TYPE_CHECKING
from uuid import uuid4

from betty.cache.memory import MemoryCache
from betty.concurrent import ensure_manager

if TYPE_CHECKING:
    from multiprocessing.managers import SyncManager
    from betty.cache import Cache


class Context:
    """
    Define a job context.
    """

    def __init__(self, *, manager: SyncManager | None = None):
        manager = ensure_manager(manager)
        self._id = str(uuid4())
        self._cache: Cache[Any] = MemoryCache(manager=manager)
        self._start = datetime.now()

    @property
    def id(self) -> str:
        """
        The unique job context ID.
        """
        return self._id

    @property
    def cache(self) -> Cache[Any]:
        """
        Provide a cache for this job context.

        The cache is volatile and will be discarded once the job context is completed.
        """
        return self._cache

    @property
    def start(self) -> datetime:
        """
        When the job started.
        """
        return self._start



"""
Provide the Raspberry Mint theme.
"""

from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING, final

import aiofiles
from typing_extensions import override

from betty.jinja2 import (
    Jinja2Provider,
    Filters,
)
from betty.locale.localizable import static, call, plain
from betty.locale.localizer import DEFAULT_LOCALIZER
from betty.os import link_or_copy
from betty.plugin import ShorthandPluginBase
from betty.project.extension import Theme, Extension, ConfigurableExtension
from betty.project.extension._theme import jinja2_filters
from betty.project.extension._theme.search import generate_search_index
from betty.project.extension.maps import Maps
from betty.project.extension.raspberry_mint.config import RaspberryMintConfiguration
from betty.project.extension.trees import Trees
from betty.project.extension.webpack import Webpack
from betty.project.extension.webpack.build import EntryPointProvider
from betty.project.generate import GenerateSiteEvent

if TYPE_CHECKING:
    from betty.plugin import PluginIdentifier
    from betty.event_dispatcher import EventHandlerRegistry
    from collections.abc import Sequence


async def _generate_logo(event: GenerateSiteEvent) -> None:
    await link_or_copy(
        event.project.logo,
        event.project.configuration.www_directory_path
        / ("logo" + event.project.logo.suffix),
    )


_RESULT_CONTAINER_TEMPLATE = plain("""
<li class="d-flex gap-2 search-result">
    {{{ betty-search-result }}}
</li>
""")

_RESULTS_CONTAINER_TEMPLATE = call(
    lambda localizer: '<ul class="entity-list"><h3 class="h2">'
    + localizer._("Results ({{{ betty-search-results-count }}})")
    + "</h3>{{{ betty-search-results }}}</ul>"
)


async def _generate_search_index(event: GenerateSiteEvent) -> None:
    await generate_search_index(
        event.project,
        _RESULT_CONTAINER_TEMPLATE,
        _RESULTS_CONTAINER_TEMPLATE,
        job_context=event.job_context,
    )


async def _generate_webmanifest(event: GenerateSiteEvent) -> None:
    project = event.project
    extensions = await project.extensions
    webmanifest = json.dumps(
        {
            "name": project.configuration.title.localize(DEFAULT_LOCALIZER),
            "icons": [
                {"src": "/logo" + project.logo.suffix},
            ],
            "lang": project.configuration.locales.default.locale,
            "theme_color": extensions[RaspberryMint].configuration.secondary_color.hex,
            "background_color": "#ffffff",
            "display": "fullscreen",
        }
    )
    async with aiofiles.open(
        project.configuration.www_directory_path / "betty.webmanifest", "w"
    ) as f:
        await f.write(webmanifest)


@final
class RaspberryMint(
    ShorthandPluginBase,
    Theme,
    ConfigurableExtension[RaspberryMintConfiguration],
    Jinja2Provider,
    EntryPointProvider,
):
    """
    The Raspberry Mint theme.
    """

    _plugin_id = "raspberry-mint"
    _plugin_label = static("Raspberry Mint")

    @override
    async def bootstrap(self) -> None:
        await super().bootstrap()
        try:
            await self._assert_configuration()
        except BaseException:
            await self.shutdown()
            raise

    async def _assert_configuration(self) -> None:
        await self.configuration.featured_entities.validate(
            self.project.entity_type_repository
        )

    @override
    def register_event_handlers(self, registry: EventHandlerRegistry) -> None:
        registry.add_handler(
            GenerateSiteEvent,
            _generate_logo,
            _generate_search_index,
            _generate_webmanifest,
        )

    @override
    @classmethod
    def depends_on(cls) -> set[PluginIdentifier[Extension]]:
        return {Webpack}

    @override
    @classmethod
    def comes_before(cls) -> set[PluginIdentifier[Extension]]:
        return {Maps, Trees}

    @override
    @classmethod
    def assets_directory_path(cls) -> Path:
        return Path(__file__).parent / "assets"

    @override
    @classmethod
    def webpack_entry_point_directory_path(cls) -> Path:
        return Path(__file__).parent / "webpack"

    @override
    def webpack_entry_point_cache_keys(self) -> Sequence[str]:
        return (
            self.project.configuration.root_path,
            self._configuration.primary_color.hex,
            self._configuration.secondary_color.hex,
            self._configuration.tertiary_color.hex,
        )

    @override
    @classmethod
    def new_default_configuration(cls) -> RaspberryMintConfiguration:
        return RaspberryMintConfiguration()

    @override
    @property
    def filters(self) -> Filters:
        return jinja2_filters(self._project)



"""
Provide no-op caching.
"""

from __future__ import annotations

from contextlib import asynccontextmanager
from typing import (
    Self,
    Any,
    AsyncContextManager,
    TypeAlias,
    overload,
    Literal,
    TYPE_CHECKING,
    final,
)

from typing_extensions import override

from betty.cache import CacheItem, Cache, CacheItemValueSetter
from betty.typing import processsafe

if TYPE_CHECKING:
    from types import TracebackType
    from collections.abc import AsyncIterator

_GetSet: TypeAlias = tuple[
    CacheItem[Any] | None,
    CacheItemValueSetter[Any] | None,
]


class _NoOpGetSet:
    def __init__(self, has_setter: bool):
        self._has_setter = has_setter

    async def __aenter__(self) -> _GetSet:
        return None, self._set if self._has_setter else None

    async def _set(self, value: Any) -> None:
        return

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        return


@final
@processsafe
class NoOpCache(Cache[Any]):
    """
    Provide a cache that does nothing.
    """

    @override
    def with_scope(self, scope: str) -> Self:
        return self

    @override
    @asynccontextmanager
    async def get(self, cache_item_id: str) -> AsyncIterator[CacheItem[Any] | None]:
        yield None

    @override
    async def set(
        self,
        cache_item_id: str,
        value: Any,
        *,
        modified: int | float | None = None,
    ) -> None:
        return

    @overload
    def getset(
        self, cache_item_id: str
    ) -> AsyncContextManager[
        tuple[
            CacheItem[Any] | None,
            CacheItemValueSetter[Any],
        ]
    ]:
        pass

    @overload
    def getset(
        self, cache_item_id: str, *, wait: Literal[False] = False
    ) -> AsyncContextManager[
        tuple[
            CacheItem[Any] | None,
            CacheItemValueSetter[Any] | None,
        ]
    ]:
        pass

    @override
    def getset(
        self, cache_item_id: str, *, wait: bool = True
    ) -> AsyncContextManager[_GetSet]:
        return _NoOpGetSet(wait)

    @override
    async def delete(self, cache_item_id: str) -> None:
        return

    @override
    async def clear(self) -> None:
        return



"""
Manipulate images.
"""

import math
from pathlib import Path
from typing import TypeAlias

from PIL import UnidentifiedImageError
from PIL.Image import Image, EXTENSION, preinit, init

from betty.media_type import MediaType
from betty.media_type.media_types import PDF

Percentage: TypeAlias = int
Pixel: TypeAlias = int
OneDimensionalSize: TypeAlias = tuple[Pixel, None] | tuple[None, Pixel]
TwoDimensionalSize: TypeAlias = tuple[Pixel, Pixel]
Size: TypeAlias = OneDimensionalSize | TwoDimensionalSize
FocusArea: TypeAlias = tuple[Percentage, Percentage, Percentage, Percentage]


def is_supported_media_type(media_type: MediaType) -> bool:
    """
    Test if a media type is supported by the image API.
    """
    if media_type.type == "image":
        return True
    if media_type == PDF:
        return True
    return False


def image_file_path_format(image_file_path: Path) -> str:
    """
    Get the PIL image format for an image's file path.
    """
    if image_file_path.suffix not in EXTENSION:
        preinit()
        if image_file_path.suffix not in EXTENSION:
            init()
            if image_file_path.suffix not in EXTENSION:
                raise UnidentifiedImageError(
                    f"cannot identify image file {image_file_path}"
                )
    return EXTENSION[image_file_path.suffix]


def _assert_size(size: Size) -> None:
    if size[0] is not None and size[0] <= 0 or size[1] is not None and size[1] <= 0:
        raise ValueError("Invalid size: sizes must be greater than 0")


def _assert_area(area: FocusArea) -> None:
    if area[0] >= area[2] or area[1] >= area[3]:
        raise ValueError("Invalid area")


def _assert_boundaries(image: Image, area: FocusArea) -> None:
    if (
        0 > area[0] > image.width
        or 0 > area[2] > image.width
        or 0 > area[1] > image.height
        or 0 > area[3] > image.height
    ):
        raise ValueError("Given area does not fit within the image.")


def resize_cover(
    original_image: Image, resize_size: Size, *, focus: FocusArea | None = None
) -> Image:
    """
    Resize an image to cover an area.

    :arg focus: An area within the image of which as much as possible should be part of the resized image.
    """
    _assert_size(resize_size)
    if focus is not None:
        _assert_area(focus)
        _assert_boundaries(original_image, focus)

    resize_width, resize_height = resize_size
    focus_left = 0 if focus is None else focus[0] * original_image.width / 100
    focus_top = 0 if focus is None else focus[1] * original_image.height / 100
    focus_right = (
        original_image.width if focus is None else focus[2] * original_image.width / 100
    )
    focus_bottom = (
        original_image.height
        if focus is None
        else focus[3] * original_image.height / 100
    )
    focus_width = focus_right - focus_left
    focus_height = focus_bottom - focus_top
    focus_ratio = focus_width / focus_height

    # Bind the maximum size by the original image.
    # This ensures the resized image won't have empty bars.
    max_width: float = original_image.width
    max_height: float = original_image.height
    # Bind the minimum size by the requested resize area, as long as the maximum allows.
    # This ensures we use at least as many pixels as needed for the resize area, reducing
    # the likelihood of the resulting image being *enlarged*.
    min_width: float = 0 if resize_width is None else min(max_width, resize_width)
    min_height: float = 0 if resize_height is None else min(max_height, resize_height)

    max_ratio = max_width / max_height
    # Adjust the minimum and maximum dimensions if the
    # resize area has a ratio (a width *and* a height).
    if resize_width is not None and resize_height is not None:
        resize_ratio = resize_width / resize_height

        # Adjust the maximum dimensions.
        if resize_ratio > max_ratio:
            # The resize area is more landscape than the original image.
            # The height is the constrained dimension.
            max_height = min(max_height, max_width / resize_ratio)
        else:
            # The resize's area is more portrait than the original image.
            # The width is the constrained dimension.
            max_width = min(max_width, max_height * resize_ratio)

        # Adjust the minimum dimensions.
        if resize_ratio > focus_ratio:
            # The resize area is more landscape than the focus area.
            # The height is the constrained dimension.
            min_height = max(min_height, focus_width / resize_ratio)
            crop_height = min(max_height, max(min_height, focus_height))
            crop_width = crop_height * resize_ratio
        else:
            # The resize's area is more portrait than the focus area.
            # The width is the constrained dimension.
            min_width = max(min_width, focus_height * resize_ratio)
            crop_width = min(max_width, max(min_width, focus_width))
            crop_height = crop_width / resize_ratio
    elif resize_width is not None:
        crop_width = max(min_width, focus_width)
        crop_height = max_height
    else:
        crop_width = max_width
        crop_height = max(min_height, focus_height)

    crop_focus_width_diff = crop_width - focus_width
    crop_focus_height_diff = crop_height - focus_height
    ideal_margin_left = math.ceil(crop_focus_width_diff / 2)
    ideal_margin_top = math.ceil(crop_focus_height_diff / 2)
    ideal_margin_right = crop_focus_width_diff - ideal_margin_left
    ideal_margin_bottom = crop_focus_height_diff - ideal_margin_top

    # Place the crop area centrally on top of the focus area.
    # At this point the crop area may still be out of bounds.
    crop_left = math.ceil(focus_left - ideal_margin_left)
    crop_top = math.ceil(focus_top - ideal_margin_top)
    crop_right = math.ceil(focus_right + ideal_margin_right)
    crop_bottom = math.ceil(focus_bottom + ideal_margin_bottom)

    # Ensure the crop area lies within the original image bounds.
    if crop_left < 0:
        crop_right -= crop_left
        crop_left = 0
    if crop_top < 0:
        crop_bottom -= crop_top
        crop_top = 0
    if crop_right > original_image.width:
        crop_left -= crop_right - original_image.width
        crop_right = original_image.width
    if crop_bottom > original_image.height:
        crop_top -= crop_bottom - original_image.height
        crop_bottom = original_image.height

    return original_image.crop((crop_left, crop_top, crop_right, crop_bottom)).resize(
        (resize_width or original_image.width, resize_height or original_image.height)
    )



"""
Provide configuration for the Cotton Candy extension.
"""

from __future__ import annotations

from typing import Sequence, TYPE_CHECKING, Iterable

from typing_extensions import override

from betty.assertion import assert_record, OptionalField
from betty.config import Configuration
from betty.model.config import EntityReference, EntityReferenceSequence
from betty.project.extension._theme import ColorConfiguration

if TYPE_CHECKING:
    from betty.mutability import Mutable
    from betty.serde.dump import Dump, DumpMapping
    from betty.model import UserFacingEntity, Entity


class CottonCandyConfiguration(Configuration):
    """
    Provide configuration for the :py:class:`betty.project.extension.cotton_candy.CottonCandy` extension.
    """

    DEFAULT_PRIMARY_INACTIVE_COLOR = "#ffc0cb"
    DEFAULT_PRIMARY_ACTIVE_COLOR = "#ff69b4"
    DEFAULT_LINK_INACTIVE_COLOR = "#149988"
    DEFAULT_LINK_ACTIVE_COLOR = "#2a615a"

    def __init__(
        self,
        *,
        featured_entities: (
            Sequence[EntityReference[UserFacingEntity & Entity]] | None
        ) = None,
        primary_inactive_color: str = DEFAULT_PRIMARY_INACTIVE_COLOR,
        primary_active_color: str = DEFAULT_PRIMARY_ACTIVE_COLOR,
        link_inactive_color: str = DEFAULT_LINK_INACTIVE_COLOR,
        link_active_color: str = DEFAULT_LINK_ACTIVE_COLOR,
    ):
        super().__init__()
        self._featured_entities = EntityReferenceSequence["UserFacingEntity & Entity"](
            featured_entities or ()
        )
        self._primary_inactive_color = ColorConfiguration(primary_inactive_color)
        self._primary_active_color = ColorConfiguration(primary_active_color)
        self._link_inactive_color = ColorConfiguration(link_inactive_color)
        self._link_active_color = ColorConfiguration(link_active_color)

    @override
    def get_mutable_instances(self) -> Iterable[Mutable]:
        return (
            self._featured_entities,
            self._primary_active_color,
            self._primary_inactive_color,
            self._link_active_color,
            self._link_inactive_color,
        )

    @property
    def featured_entities(self) -> EntityReferenceSequence[UserFacingEntity & Entity]:
        """
        The entities featured on the front page.
        """
        return self._featured_entities

    @property
    def primary_inactive_color(self) -> ColorConfiguration:
        """
        The color for inactive primary/CTA elements.
        """
        return self._primary_inactive_color

    @property
    def primary_active_color(self) -> ColorConfiguration:
        """
        The color for active primary/CTA elements.
        """
        return self._primary_active_color

    @property
    def link_inactive_color(self) -> ColorConfiguration:
        """
        The color for inactive hyperlinks.
        """
        return self._link_inactive_color

    @property
    def link_active_color(self) -> ColorConfiguration:
        """
        The color for active hyperlinks.
        """
        return self._link_active_color

    @override
    def load(self, dump: Dump) -> None:
        self.assert_mutable()
        assert_record(
            OptionalField("featured_entities", self.featured_entities.load),
            OptionalField("primary_inactive_color", self.primary_inactive_color.load),
            OptionalField("primary_active_color", self.primary_active_color.load),
            OptionalField("link_inactive_color", self.link_inactive_color.load),
            OptionalField("link_active_color", self.link_active_color.load),
        )(dump)

    @override
    def dump(self) -> DumpMapping[Dump]:
        return {
            "featured_entities": self.featured_entities.dump(),
            "primary_inactive_color": self._primary_inactive_color.dump(),
            "primary_active_color": self._primary_active_color.dump(),
            "link_inactive_color": self._link_inactive_color.dump(),
            "link_active_color": self._link_active_color.dump(),
        }



"""
Provide application configuration.
"""

from __future__ import annotations

from typing import final, TYPE_CHECKING

from typing_extensions import override

from betty import fs
from betty.assertion import (
    assert_record,
    OptionalField,
    assert_str,
    assert_setattr,
    assert_locale,
)
from betty.config import Configuration
from betty.typing import pickleable

if TYPE_CHECKING:
    from betty.serde.dump import Dump, DumpMapping

CONFIGURATION_FILE_PATH = fs.HOME_DIRECTORY_PATH / "app.json"


@final
@pickleable
class AppConfiguration(Configuration):
    """
    Provide configuration for :py:class:`betty.app.App`.
    """

    def __init__(
        self,
        *,
        locale: str | None = None,
    ):
        super().__init__()
        self._locale: str | None = locale

    @property
    def locale(self) -> str | None:
        """
        The application locale.
        """
        return self._locale

    @locale.setter
    def locale(self, locale: str) -> None:
        self.assert_mutable()
        self._locale = assert_locale()(locale)

    @override
    def load(self, dump: Dump) -> None:
        self.assert_mutable()
        assert_record(
            OptionalField("locale", assert_str() | assert_setattr(self, "locale"))
        )(dump)

    @override
    def dump(self) -> DumpMapping[Dump]:
        if self.locale is None:
            return {}
        return {"locale": self.locale}



"""
Provide demonstration site functionality.
"""

from __future__ import annotations

from asyncio import to_thread
from contextlib import suppress
from shutil import rmtree
from typing import TYPE_CHECKING, final, Sequence

from typing_extensions import override

from betty.html import NavigationLinkProvider, NavigationLink
from betty.locale.localizable import static, _
from betty.plugin import ShorthandPluginBase
from betty.project import generate
from betty.project.extension import Extension
from betty.project.extension.demo.project import load_ancestry
from betty.project.extension.deriver import Deriver
from betty.project.extension.http_api_doc import HttpApiDoc
from betty.project.extension.maps import Maps
from betty.project.extension.raspberry_mint import RaspberryMint
from betty.project.extension.trees import Trees
from betty.project.extension.wikipedia import Wikipedia
from betty.project.load import LoadAncestryEvent, load
from betty.typing import internal

if TYPE_CHECKING:
    from betty.project import Project
    from betty.plugin import PluginIdentifier
    from betty.event_dispatcher import EventHandlerRegistry


@internal
async def generate_with_cleanup(project: Project) -> None:
    """
    Generate a demonstration site, and clean up the project directory on any errors.
    """
    if project.configuration.www_directory_path.exists():
        return
    await load(project)
    with suppress(FileNotFoundError):
        await to_thread(rmtree, project.configuration.project_directory_path)
    try:
        await generate.generate(project)
    except BaseException:
        with suppress(FileNotFoundError):
            await to_thread(rmtree, project.configuration.project_directory_path)
        raise


@final
class Demo(ShorthandPluginBase, NavigationLinkProvider, Extension):
    """
    Provide demonstration site functionality.
    """

    _plugin_id = "demo"
    _plugin_label = static("Demo")

    @override
    @classmethod
    def depends_on(cls) -> set[PluginIdentifier[Extension]]:
        return {
            Deriver,
            HttpApiDoc,
            Maps,
            RaspberryMint,
            Trees,
            Wikipedia,
        }

    @override
    def register_event_handlers(self, registry: EventHandlerRegistry) -> None:
        registry.add_handler(
            LoadAncestryEvent, lambda event: load_ancestry(event.project)
        )

    @override
    def secondary_navigation_links(self) -> Sequence[NavigationLink]:
        return [
            NavigationLink(
                "https://github.com/bartfeenstra/betty", _("Find Betty on GitHub")
            ),
            NavigationLink(
                "https://betty.readthedocs.io/", _("Read the Betty documentation")
            ),
        ]



"""
Data types that have a locale.
"""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from typing_extensions import override

from betty.json.linked_data import LinkedDataDumpableJsonLdObject, JsonLdObject
from betty.json.schema import Null, OneOf
from betty.locale import UNDETERMINED_LOCALE, LocaleSchema
from betty.locale.localized import Localized
from betty.privacy import is_public

if TYPE_CHECKING:
    from betty.serde.dump import DumpMapping, Dump
    from betty.project import Project


class HasLocale(Localized, LinkedDataDumpableJsonLdObject):
    """
    A resource that is localized, e.g. contains information in a specific locale.
    """

    def __init__(
        self,
        *args: Any,
        locale: str = UNDETERMINED_LOCALE,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._locale = locale

    @override
    @property
    def locale(self) -> str:
        return self._locale

    @locale.setter
    def locale(self, locale: str) -> None:
        self._locale = locale

    @override
    async def dump_linked_data(self, project: Project) -> DumpMapping[Dump]:
        dump = await super().dump_linked_data(project)
        dump["locale"] = self.locale if is_public(self) else None
        return dump

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property("locale", OneOf(LocaleSchema(), Null()))
        return schema



"""
Provide Betty's ancestry event types.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from typing_extensions import override

from betty.ancestry.event_type import EventType
from betty.locale.localizable import _
from betty.locale.localizer import DEFAULT_LOCALIZER
from betty.plugin import ShorthandPluginBase, PluginIdentifier

if TYPE_CHECKING:
    from betty.ancestry.person import Person


@final
class Unknown(ShorthandPluginBase, EventType):
    """
    Describe an event for which no more specific type is known.
    """

    _plugin_id = "unknown"
    _plugin_label = _("Unknown")


class DerivableEventType(EventType):
    """
    Any event that that may be updated by the deriver API.
    """

    pass  # pragma: no cover


class CreatableDerivableEventType(DerivableEventType):
    """
    Any event type of which the deriver API may create new instances.
    """

    @classmethod
    def may_create(cls, person: Person, lifetime_threshold: int) -> bool:
        """
        Whether a new event of this type may be created for the given person.
        """
        return True  # pragma: no cover


class PreBirthEventType(EventType):
    """
    Any event that only ever takes place before someone's birth.
    """

    @override
    @classmethod
    def comes_before(cls) -> set[PluginIdentifier[EventType]]:
        return {Birth}  # pragma: no cover


class StartOfLifeEventType(EventType):
    """
    An event that indicates the start of someone's life.

    This includes someone's actual birth, but also other types of events that take place
    close to someone's birth and as such are indicators that that person was born around
    the time of the start-of-life event.
    """

    pass  # pragma: no cover


class DuringLifeEventType(EventType):
    """
    Any event that only ever takes place during someone's life, e.g. after their birth and before their death.
    """

    @override
    @classmethod
    def comes_after(cls) -> set[PluginIdentifier[EventType]]:
        return {Birth}  # pragma: no cover

    @override
    @classmethod
    def comes_before(cls) -> set[PluginIdentifier[EventType]]:
        return {Death}  # pragma: no cover


class EndOfLifeEventType(EventType):
    """
    An event that indicates the end of someone's life.

    This includes someone's actual death, but also other types of events that take place
    close to someone's death and as such are indicators that that person died around the
    time of the end-of-life event.
    """

    pass  # pragma: no cover


class PostDeathEventType(EventType):
    """
    An event that only ever happens after someone's death.
    """

    @override
    @classmethod
    def comes_after(cls) -> set[PluginIdentifier[EventType]]:
        return {Death}  # pragma: no cover


@final
class Birth(CreatableDerivableEventType, StartOfLifeEventType, ShorthandPluginBase):
    """
    Someone was born.
    """

    _plugin_id = "birth"
    _plugin_label = _("Birth")

    @override
    @classmethod
    def comes_before(cls) -> set[PluginIdentifier[EventType]]:
        return {DuringLifeEventType}  # pragma: no cover


@final
class Baptism(DuringLifeEventType, StartOfLifeEventType, ShorthandPluginBase):
    """
    Someone was `baptized <https://en.wikipedia.org/wiki/Baptism>`_.
    """

    _plugin_id = "baptism"
    _plugin_label = _("Baptism")


@final
class BarMitzvah(DuringLifeEventType, StartOfLifeEventType, ShorthandPluginBase):
    """
    Someone's `bar mitzvah <https://en.wikipedia.org/wiki/Bar_and_bat_mitzvah>`_ took place.
    """

    _plugin_id = "bar-mitzvah"
    _plugin_label = _("Bar mitzvah")


@final
class BatMitzvah(DuringLifeEventType, StartOfLifeEventType, ShorthandPluginBase):
    """
    Someone's `bat mitzvah <https://en.wikipedia.org/wiki/Bar_and_bat_mitzvah>`_ took place.
    """

    _plugin_id = "bat-mitzvah"
    _plugin_label = _("Bat mitzvah")


@final
class Adoption(DuringLifeEventType, ShorthandPluginBase):
    """
    Someone was adopted.
    """

    _plugin_id = "adoption"
    _plugin_label = _("Adoption")


@final
class Death(CreatableDerivableEventType, EndOfLifeEventType, ShorthandPluginBase):
    """
    Someone died.
    """

    _plugin_id = "death"
    _plugin_label = _("Death")

    @override
    @classmethod
    def comes_after(cls) -> set[PluginIdentifier[EventType]]:
        return {DuringLifeEventType}  # pragma: no cover

    @override
    @classmethod
    def may_create(cls, person: Person, lifetime_threshold: int) -> bool:
        from betty.privacy.privatizer import Privatizer

        return Privatizer(lifetime_threshold, localizer=DEFAULT_LOCALIZER).has_expired(
            person, 1
        )


class FinalDispositionEventType(
    PostDeathEventType, DerivableEventType, EndOfLifeEventType
):
    """
    Someone's `final disposition <https://en.wikipedia.org/wiki/Disposal_of_human_corpses>`_ took place.
    """

    pass  # pragma: no cover


@final
class Funeral(FinalDispositionEventType, ShorthandPluginBase):
    """
    Someone's funeral took place.
    """

    _plugin_id = "funeral"
    _plugin_label = _("Funeral")


@final
class Cremation(FinalDispositionEventType, ShorthandPluginBase):
    """
    Someone was cremated.
    """

    _plugin_id = "cremation"
    _plugin_label = _("Cremation")


@final
class Burial(FinalDispositionEventType, ShorthandPluginBase):
    """
    Someone was buried.
    """

    _plugin_id = "burial"
    _plugin_label = _("Burial")


@final
class Will(PostDeathEventType, ShorthandPluginBase):
    """
    Someone's `will and testament <https://en.wikipedia.org/wiki/Will_and_testament>`_ came into effect.
    """

    _plugin_id = "will"
    _plugin_label = _("Will")


@final
class Engagement(DuringLifeEventType, ShorthandPluginBase):
    """
    People got engaged with the intent to marry.
    """

    _plugin_id = "engagement"
    _plugin_label = _("Engagement")

    @override
    @classmethod
    def comes_before(cls) -> set[PluginIdentifier[EventType]]:
        return {Marriage}  # pragma: no cover


@final
class Marriage(DuringLifeEventType, ShorthandPluginBase):
    """
    People were married.
    """

    _plugin_id = "marriage"
    _plugin_label = _("Marriage")


@final
class MarriageAnnouncement(DuringLifeEventType, ShorthandPluginBase):
    """
    People's marriage was announced.
    """

    _plugin_id = "marriage-announcement"
    _plugin_label = _("Announcement of marriage")

    @override
    @classmethod
    def comes_before(cls) -> set[PluginIdentifier[EventType]]:
        return {Marriage}  # pragma: no cover


@final
class Divorce(DuringLifeEventType, ShorthandPluginBase):
    """
    People were divorced.
    """

    _plugin_id = "divorce"
    _plugin_label = _("Divorce")

    @override
    @classmethod
    def comes_after(cls) -> set[PluginIdentifier[EventType]]:
        return {Marriage}  # pragma: no cover


@final
class DivorceAnnouncement(DuringLifeEventType, ShorthandPluginBase):
    """
    People's divorce was announced.
    """

    _plugin_id = "divorce-announcement"
    _plugin_label = _("Announcement of divorce")

    @override
    @classmethod
    def comes_after(cls) -> set[PluginIdentifier[EventType]]:
        return {Marriage}  # pragma: no cover

    @override
    @classmethod
    def comes_before(cls) -> set[PluginIdentifier[EventType]]:
        return {Divorce}  # pragma: no cover


@final
class Residence(DuringLifeEventType, ShorthandPluginBase):
    """
    Someone resided/lived in a place.
    """

    _plugin_id = "residence"
    _plugin_label = _("Residence")


@final
class Immigration(DuringLifeEventType, ShorthandPluginBase):
    """
    Someone immigrated to a place.
    """

    _plugin_id = "immigration"
    _plugin_label = _("Immigration")


@final
class Emigration(ShorthandPluginBase, DuringLifeEventType):
    """
    Someone emigrated from a place.
    """

    _plugin_id = "emigration"
    _plugin_label = _("Emigration")


@final
class Occupation(ShorthandPluginBase, DuringLifeEventType):
    """
    Someone's occupation, e.g. their main recurring activity.

    This may include employment, education, stay at home parent, etc.
    """

    _plugin_id = "occupation"
    _plugin_label = _("Occupation")


@final
class Retirement(ShorthandPluginBase, DuringLifeEventType):
    """
    Someone `retired <https://en.wikipedia.org/wiki/Retirement>`_.
    """

    _plugin_id = "retirement"
    _plugin_label = _("Retirement")


@final
class Correspondence(ShorthandPluginBase, EventType):
    """
    People corresponded with each other.
    """

    _plugin_id = "correspondence"
    _plugin_label = _("Correspondence")


@final
class Confirmation(ShorthandPluginBase, DuringLifeEventType):
    """
    Someone's `confirmation <https://en.wikipedia.org/wiki/Confirmation>`_ took place.
    """

    _plugin_id = "confirmation"
    _plugin_label = _("Confirmation")


@final
class Missing(ShorthandPluginBase, DuringLifeEventType):
    """
    Someone went missing.
    """

    _plugin_id = "missing"
    _plugin_label = _("Missing")


@final
class Conference(ShorthandPluginBase, DuringLifeEventType):
    """
    A conference between people took place.
    """

    _plugin_id = "conference"
    _plugin_label = _("Conference")



"""
The Link API allows data to reference external resources.
"""

from __future__ import annotations

from typing import final, Any, MutableSequence, TYPE_CHECKING, Self

from typing_extensions import override

from betty.ancestry.description import HasDescription
from betty.ancestry.locale import HasLocale
from betty.ancestry.media_type import HasMediaType
from betty.json.linked_data import (
    JsonLdObject,
    dump_link,
    LinkedDataDumpableJsonLdObject,
    JsonLdSchema,
)
from betty.json.schema import String, Array
from betty.link import Link as StdLink
from betty.locale import UNDETERMINED_LOCALE
from betty.locale.localizable import (
    OptionalStaticTranslationsLocalizableAttr,
    ShorthandStaticTranslations,
    StaticTranslationsLocalizableSchema,
    StaticTranslationsLocalizable,
)
from betty.privacy import is_public

if TYPE_CHECKING:
    from betty.serde.dump import DumpMapping, Dump
    from betty.project import Project
    from betty.media_type import MediaType


@final
class Link(
    StdLink, HasMediaType, HasLocale, HasDescription, LinkedDataDumpableJsonLdObject
):
    """
    An external link.
    """

    #: The link's `IANA link relationship <https://www.iana.org/assignments/link-relations/link-relations.xhtml>`_.
    relationship: str | None
    _label = OptionalStaticTranslationsLocalizableAttr("_label", title="Label")

    def __init__(
        self,
        url: str,
        *,
        relationship: str | None = None,
        label: ShorthandStaticTranslations | None = None,
        description: ShorthandStaticTranslations | None = None,
        media_type: MediaType | None = None,
        locale: str = UNDETERMINED_LOCALE,
    ):
        super().__init__(
            media_type=media_type,
            description=description,
            locale=locale,
        )
        self._url = url
        if label:
            self._label = label
        self.relationship = relationship

    @override
    @property
    def url(self) -> str:
        return self._url

    @url.setter
    def url(self, url: str) -> None:
        self._url = url

    @override
    @property
    def label(self) -> StaticTranslationsLocalizable:
        return self._label

    @label.setter
    def label(self, label: ShorthandStaticTranslations) -> None:
        self._label = label

    @label.deleter
    def label(self) -> None:
        del self._label

    @override
    async def dump_linked_data(self, project: Project) -> DumpMapping[Dump]:
        dump = await super().dump_linked_data(project)
        dump["url"] = self.url
        if self._label:
            dump["label"] = await self._label.dump_linked_data(project)
        if self.relationship is not None:
            dump["relationship"] = self.relationship
        return dump

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project) -> LinkSchema:
        return await LinkSchema.new()


@final
class LinkSchema(JsonLdObject):
    """
    A JSON Schema for :py:class:`betty.ancestry.link.Link`.
    """

    def __init__(self, json_ld_schema: JsonLdSchema):
        super().__init__(json_ld_schema, def_name="link", title="Link")
        self.add_property(
            "url",
            String(
                format=String.Format.URI,
                description="The full URL to the other resource.",
            ),
        )
        self.add_property(
            "relationship",
            String(
                description="The relationship between this resource and the link target (https://en.wikipedia.org/wiki/Link_relation)."
            ),
            False,
        )
        self.add_property(
            "label",
            StaticTranslationsLocalizableSchema(
                title="Label", description="The human-readable link label."
            ),
            False,
        )

    @classmethod
    async def new(cls) -> Self:
        """
        Create a new instance.
        """
        return cls(await JsonLdSchema.new())


class LinkCollectionSchema(Array):
    """
    A JSON Schema for :py:class:`betty.ancestry.link.Link` collections.
    """

    def __init__(self, link_schema: LinkSchema):
        super().__init__(link_schema, def_name="linkCollection", title="Links")

    @classmethod
    async def new(cls) -> Self:
        """
        Create a new instance.
        """
        return cls(await LinkSchema.new())


class HasLinks(LinkedDataDumpableJsonLdObject):
    """
    A resource that has external links.
    """

    def __init__(
        self,
        *args: Any,
        links: MutableSequence[Link] | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._links: MutableSequence[Link] = links if links else []

    @property
    def links(self) -> MutableSequence[Link]:
        """
        The extenal links.
        """
        return self._links

    @override
    async def dump_linked_data(self, project: Project) -> DumpMapping[Dump]:
        dump = await super().dump_linked_data(project)
        await dump_link(
            dump,
            project,
            *(self.links if is_public(self) else ()),
        )

        return dump

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property("links", await LinkCollectionSchema.new())
        return schema



"""Provide interactive family trees by integrating Betty with `Cytoscape.js <https://cytoscape.org/>`_."""

from __future__ import annotations

import json
from asyncio import gather
from pathlib import Path
from typing import TYPE_CHECKING, final

import aiofiles
from typing_extensions import override

from betty.ancestry.person import Person
from betty.locale.localizable import _
from betty.media_type.media_types import HTML
from betty.plugin import ShorthandPluginBase
from betty.project.extension.webpack import Webpack
from betty.project.extension.webpack.build import EntryPointProvider
from betty.project.generate import GenerateSiteEvent

if TYPE_CHECKING:
    from betty.project.extension import Extension
    from betty.event_dispatcher import EventHandlerRegistry
    from betty.plugin import PluginIdentifier
    from collections.abc import Sequence


async def _generate_people_json(event: GenerateSiteEvent) -> None:
    await gather(
        *(
            _generate_people_json_for_locale(event, locale)
            for locale in event.project.configuration.locales
        )
    )


async def _generate_people_json_for_locale(
    event: GenerateSiteEvent, locale: str
) -> None:
    project = event.project
    url_generator = await project.url_generator
    localizers = await project.localizers
    localizer = await localizers.get(locale)
    private_label = localizer._("private")
    people = {
        person.id: {
            "id": person.id,
            "label": person.label.localize(localizer)
            if person.public
            else private_label,
            "url": url_generator.generate(person, media_type=HTML),
            "parentIds": [parent.id for parent in person.parents],
            "childIds": [child.id for child in person.children],
            "private": person.private,
        }
        for person in project.ancestry[Person]
    }
    people_json = json.dumps(people)
    async with aiofiles.open(
        project.configuration.localize_www_directory_path(locale) / "people.json",
        mode="w",
    ) as f:
        await f.write(people_json)


@final
class Trees(ShorthandPluginBase, EntryPointProvider):
    """
    Provide interactive family trees for use in web pages.
    """

    _plugin_id = "trees"
    _plugin_label = _("Trees")
    _plugin_description = _(
        'Display interactive family trees using <a href="https://cytoscape.org/">Cytoscape</a>.'
    )

    @override
    @classmethod
    def depends_on(cls) -> set[PluginIdentifier[Extension]]:
        return {Webpack}

    @override
    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return Path(__file__).parent / "assets"

    @override
    def register_event_handlers(self, registry: EventHandlerRegistry) -> None:
        registry.add_handler(GenerateSiteEvent, _generate_people_json)

    @override
    @classmethod
    def webpack_entry_point_directory_path(cls) -> Path:
        return Path(__file__).parent / "webpack"

    @override
    def webpack_entry_point_cache_keys(self) -> Sequence[str]:
        return ()



"""
Provide Betty's ancestry genders.
"""

from __future__ import annotations

from betty.mutability import Mutable
from betty.plugin import Plugin, PluginRepository
from betty.plugin.entry_point import EntryPointPluginRepository


class Gender(Mutable, Plugin):
    """
    Define a gender.

    Read more about :doc:`/development/plugin/gender`.

    To test your own subclasses, use :py:class:`betty.test_utils.ancestry.gender.GenderTestBase`.
    """

    pass


GENDER_REPOSITORY: PluginRepository[Gender] = EntryPointPluginRepository("betty.gender")
"""
The gender plugin repository.

Read more about :doc:`/development/plugin/gender`.
"""
